"""
Manage VMware ESXi Hosts.

.. Warning::
    This module will be deprecated in a future release of Salt. VMware strongly
    recommends using the
    `VMware Salt extensions <https://docs.saltproject.io/salt/extensions/salt-ext-modules-vmware/en/latest/all.html>`_
    instead of the ESXi module. Because the Salt extensions are newer and
    actively supported by VMware, they are more compatible with current versions
    of ESXi and they work well with the latest features in the VMware product
    line.


Dependencies
============

- pyVmomi Python Module
- ESXCLI


pyVmomi
-------

PyVmomi can be installed via pip:

.. code-block:: bash

    pip install pyVmomi

.. note::

    Version 6.0 of pyVmomi has some problems with SSL error handling on certain
    versions of Python. If using version 6.0 of pyVmomi, Python 2.6,
    Python 2.7.9, or newer must be present. This is due to an upstream dependency
    in pyVmomi 6.0 that is not supported in Python versions 2.7 to 2.7.8. If the
    version of Python is not in the supported range, you will need to install an
    earlier version of pyVmomi. See `Issue #29537`_ for more information.

.. _Issue #29537: https://github.com/saltstack/salt/issues/29537

Based on the note above, to install an earlier version of pyVmomi than the
version currently listed in PyPi, run the following:

.. code-block:: bash

    pip install pyVmomi==5.5.0.2014.1.1

The 5.5.0.2014.1.1 is a known stable version that this original ESXi State
Module was developed against.

ESXCLI
------

Currently, about a third of the functions used in the vSphere Execution Module require
the ESXCLI package be installed on the machine running the Proxy Minion process.

The ESXCLI package is also referred to as the VMware vSphere CLI, or vCLI. VMware
provides vCLI package installation instructions for `vSphere 5.5`_ and
`vSphere 6.0`_.

.. _vSphere 5.5: http://pubs.vmware.com/vsphere-55/index.jsp#com.vmware.vcli.getstart.doc/cli_install.4.2.html
.. _vSphere 6.0: http://pubs.vmware.com/vsphere-60/index.jsp#com.vmware.vcli.getstart.doc/cli_install.4.2.html

Once all of the required dependencies are in place and the vCLI package is
installed, you can check to see if you can connect to your ESXi host or vCenter
server by running the following command:

.. code-block:: bash

    esxcli -s <host-location> -u <username> -p <password> system syslog config get

If the connection was successful, ESXCLI was successfully installed on your system.
You should see output related to the ESXi host's syslog configuration.

.. note::

    Be aware that some functionality in this state module may depend on the
    type of license attached to the ESXi host.

    For example, certain services are only available to manipulate service state
    or policies with a VMware vSphere Enterprise or Enterprise Plus license, while
    others are available with a Standard license. The ``ntpd`` service is restricted
    to an Enterprise Plus license, while ``ssh`` is available via the Standard
    license.

    Please see the `vSphere Comparison`_ page for more information.

.. _vSphere Comparison: https://www.vmware.com/products/vsphere/compare

About
-----

This state module was written to be used in conjunction with Salt's
:mod:`ESXi Proxy Minion <salt.proxy.esxi>`. For a tutorial on how to use Salt's
ESXi Proxy Minion, please refer to the
:ref:`ESXi Proxy Minion Tutorial <tutorial-esxi-proxy>` for
configuration examples, dependency installation instructions, how to run remote
execution functions against ESXi hosts via a Salt Proxy Minion, and a larger state
example.
"""

import logging
import re
import sys
from functools import wraps

import salt.utils.files
from salt.config.schemas.esxi import DiskGroupsDiskScsiAddressSchema, HostCacheSchema
from salt.exceptions import (
    ArgumentValueError,
    CommandExecutionError,
    InvalidConfigError,
    VMwareApiError,
    VMwareObjectRetrievalError,
    VMwareSaltError,
)
from salt.utils.decorators import depends

# External libraries
try:
    import jsonschema

    HAS_JSONSCHEMA = True
except ImportError:
    HAS_JSONSCHEMA = False

# Get Logging Started
log = logging.getLogger(__name__)

try:
    from pyVmomi import VmomiSupport

    # We check the supported vim versions to infer the pyVmomi version
    if (
        "vim25/6.0" in VmomiSupport.versionMap
        and sys.version_info > (2, 7)
        and sys.version_info < (2, 7, 9)
    ):

        log.debug(
            "pyVmomi not loaded: Incompatible versions of Python. See Issue #29537."
        )
        raise ImportError()
    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False


def __virtual__():
    if "esxi.cmd" in __salt__:
        return True
    return (False, "esxi module could not be loaded")


def _deprecation_message(function):
    """
    Decorator wrapper to warn about azurearm deprecation
    """

    @wraps(function)
    def wrapped(*args, **kwargs):
        salt.utils.versions.warn_until(
            3008,
            "The 'esxi' functionality in Salt has been deprecated and its "
            "functionality will be removed in version 3008 in favor of the "
            "saltext.vmware Salt Extension. "
            "(https://github.com/saltstack/salt-ext-modules-vmware)",
            category=FutureWarning,
        )
        ret = function(*args, **salt.utils.args.clean_kwargs(**kwargs))
        return ret

    return wrapped


@_deprecation_message
def coredump_configured(name, enabled, dump_ip, host_vnic="vmk0", dump_port=6500):
    """
    Ensures a host's core dump configuration.

    name
        Name of the state.

    enabled
        Sets whether or not ESXi core dump collection should be enabled.
        This is a boolean value set to ``True`` or ``False`` to enable
        or disable core dumps.

        Note that ESXi requires that the core dump must be enabled before
        any other parameters may be set. This also affects the ``changes``
        results in the state return dictionary. If ``enabled`` is ``False``,
        we can't obtain any previous settings to compare other state variables,
        resulting in many ``old`` references returning ``None``.

        Once ``enabled`` is ``True`` the ``changes`` dictionary comparisons
        will be more accurate. This is due to the way the system coredemp
        network configuration command returns data.

    dump_ip
        The IP address of host that will accept the dump.

    host_vnic
        Host VNic port through which to communicate. Defaults to ``vmk0``.

    dump_port
        TCP port to use for the dump. Defaults to ``6500``.

    Example:

    .. code-block:: yaml

        configure-host-coredump:
          esxi.coredump_configured:
            - enabled: True
            - dump_ip: 'my-coredump-ip.example.com'

    """
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    esxi_cmd = "esxi.cmd"
    enabled_msg = (
        "ESXi requires that the core dump must be enabled "
        "before any other parameters may be set."
    )
    host = __pillar__["proxy"]["host"]

    current_config = __salt__[esxi_cmd]("get_coredump_network_config").get(host)
    error = current_config.get("Error")
    if error:
        ret["comment"] = f"Error: {error}"
        return ret

    current_config = current_config.get("Coredump Config")
    current_enabled = current_config.get("enabled")

    # Configure coredump enabled state, if there are changes.
    if current_enabled != enabled:
        enabled_changes = {"enabled": {"old": current_enabled, "new": enabled}}
        # Only run the command if not using test=True
        if not __opts__["test"]:
            response = __salt__[esxi_cmd](
                "coredump_network_enable", enabled=enabled
            ).get(host)
            error = response.get("Error")
            if error:
                ret["comment"] = f"Error: {error}"
                return ret

            # Allow users to disable core dump, but then return since
            # nothing else can be set if core dump is disabled.
            if not enabled:
                ret["result"] = True
                ret["comment"] = enabled_msg
                ret["changes"].update(enabled_changes)
                return ret

        ret["changes"].update(enabled_changes)

    elif not enabled:
        # If current_enabled and enabled match, but are both False,
        # We must return before configuring anything. This isn't a
        # failure as core dump may be disabled intentionally.
        ret["result"] = True
        ret["comment"] = enabled_msg
        return ret

    # Test for changes with all remaining configurations. The changes flag is used
    # To detect changes, and then set_coredump_network_config is called one time.
    changes = False
    current_ip = current_config.get("ip")
    if current_ip != dump_ip:
        ret["changes"].update({"dump_ip": {"old": current_ip, "new": dump_ip}})
        changes = True

    current_vnic = current_config.get("host_vnic")
    if current_vnic != host_vnic:
        ret["changes"].update({"host_vnic": {"old": current_vnic, "new": host_vnic}})
        changes = True

    current_port = current_config.get("port")
    if current_port != str(dump_port):
        ret["changes"].update(
            {"dump_port": {"old": current_port, "new": str(dump_port)}}
        )
        changes = True

    # Only run the command if not using test=True and changes were detected.
    if not __opts__["test"] and changes is True:
        response = __salt__[esxi_cmd](
            "set_coredump_network_config",
            dump_ip=dump_ip,
            host_vnic=host_vnic,
            dump_port=dump_port,
        ).get(host)
        if response.get("success") is False:
            msg = response.get("stderr")
            if not msg:
                msg = response.get("stdout")
            ret["comment"] = f"Error: {msg}"
            return ret

    ret["result"] = True
    if ret["changes"] == {}:
        ret["comment"] = "Core Dump configuration is already in the desired state."
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Core dump configuration will change."

    return ret


@_deprecation_message
def password_present(name, password):
    """
    Ensures the given password is set on the ESXi host. Passwords cannot be obtained from
    host, so if a password is set in this state, the ``vsphere.update_host_password``
    function will always run (except when using test=True functionality) and the state's
    changes dictionary will always be populated.

    The username for which the password will change is the same username that is used to
    authenticate against the ESXi host via the Proxy Minion. For example, if the pillar
    definition for the proxy username is defined as ``root``, then the username that the
    password will be updated for via this state is ``root``.

    name
        Name of the state.

    password
        The new password to change on the host.

    Example:

    .. code-block:: yaml

        configure-host-password:
          esxi.password_present:
            - password: 'new-bad-password'
    """
    ret = {
        "name": name,
        "result": True,
        "changes": {"old": "unknown", "new": "********"},
        "comment": "Host password was updated.",
    }
    esxi_cmd = "esxi.cmd"

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Host password will change."
        return ret
    else:
        try:
            __salt__[esxi_cmd]("update_host_password", new_password=password)
        except CommandExecutionError as err:
            ret["result"] = False
            ret["comment"] = f"Error: {err}"
            return ret

    return ret


@_deprecation_message
def ntp_configured(
    name,
    service_running,
    ntp_servers=None,
    service_policy=None,
    service_restart=False,
    update_datetime=False,
):
    """
    Ensures a host's NTP server configuration such as setting NTP servers, ensuring the
    NTP daemon is running or stopped, or restarting the NTP daemon for the ESXi host.

    name
        Name of the state.

    service_running
        Ensures the running state of the ntp daemon for the host. Boolean value where
        ``True`` indicates that ntpd should be running and ``False`` indicates that it
        should be stopped.

    ntp_servers
        A list of servers that should be added to the ESXi host's NTP configuration.

    service_policy
        The policy to set for the NTP service.

        .. note::

            When setting the service policy to ``off`` or ``on``, you *must* quote the
            setting. If you don't, the yaml parser will set the string to a boolean,
            which will cause trouble checking for stateful changes and will error when
            trying to set the policy on the ESXi host.


    service_restart
        If set to ``True``, the ntp daemon will be restarted, regardless of its previous
        running state. Default is ``False``.

    update_datetime
        If set to ``True``, the date/time on the given host will be updated to UTC.
        Default setting is ``False``. This option should be used with caution since
        network delays and execution delays can result in time skews.

    Example:

    .. code-block:: yaml

        configure-host-ntp:
          esxi.ntp_configured:
            - service_running: True
            - ntp_servers:
              - 192.174.1.100
              - 192.174.1.200
            - service_policy: 'on'
            - service_restart: True

    """
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    esxi_cmd = "esxi.cmd"
    host = __pillar__["proxy"]["host"]
    ntpd = "ntpd"

    ntp_config = __salt__[esxi_cmd]("get_ntp_config").get(host)
    ntp_running = __salt__[esxi_cmd]("get_service_running", service_name=ntpd).get(host)
    error = ntp_running.get("Error")
    if error:
        ret["comment"] = f"Error: {error}"
        return ret
    ntp_running = ntp_running.get(ntpd)

    # Configure NTP Servers for the Host
    if ntp_servers and set(ntp_servers) != set(ntp_config):
        # Only run the command if not using test=True
        if not __opts__["test"]:
            response = __salt__[esxi_cmd](
                "set_ntp_config", ntp_servers=ntp_servers
            ).get(host)
            error = response.get("Error")
            if error:
                ret["comment"] = f"Error: {error}"
                return ret
        # Set changes dictionary for ntp_servers
        ret["changes"].update({"ntp_servers": {"old": ntp_config, "new": ntp_servers}})

    # Configure service_running state
    if service_running != ntp_running:
        # Only run the command if not using test=True
        if not __opts__["test"]:
            # Start ntdp if service_running=True
            if ntp_running is True:
                response = __salt__[esxi_cmd]("service_start", service_name=ntpd).get(
                    host
                )
                error = response.get("Error")
                if error:
                    ret["comment"] = f"Error: {error}"
                    return ret
            # Stop ntpd if service_running=False
            else:
                response = __salt__[esxi_cmd]("service_stop", service_name=ntpd).get(
                    host
                )
                error = response.get("Error")
                if error:
                    ret["comment"] = f"Error: {error}"
                    return ret
        ret["changes"].update(
            {"service_running": {"old": ntp_running, "new": service_running}}
        )

    # Configure service_policy
    if service_policy:
        current_service_policy = __salt__[esxi_cmd](
            "get_service_policy", service_name=ntpd
        ).get(host)
        error = current_service_policy.get("Error")
        if error:
            ret["comment"] = f"Error: {error}"
            return ret
        current_service_policy = current_service_policy.get(ntpd)

        if service_policy != current_service_policy:
            # Only run the command if not using test=True
            if not __opts__["test"]:
                response = __salt__[esxi_cmd](
                    "set_service_policy",
                    service_name=ntpd,
                    service_policy=service_policy,
                ).get(host)
                error = response.get("Error")
                if error:
                    ret["comment"] = f"Error: {error}"
                    return ret
            ret["changes"].update(
                {
                    "service_policy": {
                        "old": current_service_policy,
                        "new": service_policy,
                    }
                }
            )

    # Update datetime, if requested.
    if update_datetime:
        # Only run the command if not using test=True
        if not __opts__["test"]:
            response = __salt__[esxi_cmd]("update_host_datetime").get(host)
            error = response.get("Error")
            if error:
                ret["comment"] = f"Error: {error}"
                return ret
        ret["changes"].update(
            {"update_datetime": {"old": "", "new": "Host datetime was updated."}}
        )

    # Restart ntp_service if service_restart=True
    if service_restart:
        # Only run the command if not using test=True
        if not __opts__["test"]:
            response = __salt__[esxi_cmd]("service_restart", service_name=ntpd).get(
                host
            )
            error = response.get("Error")
            if error:
                ret["comment"] = f"Error: {error}"
                return ret
        ret["changes"].update(
            {"service_restart": {"old": "", "new": "NTP Daemon Restarted."}}
        )

    ret["result"] = True
    if ret["changes"] == {}:
        ret["comment"] = "NTP is already in the desired state."
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "NTP state will change."

    return ret


@_deprecation_message
def vmotion_configured(name, enabled, device="vmk0"):
    """
    Configures a host's VMotion properties such as enabling VMotion and setting
    the device VirtualNic that VMotion will use.

    name
        Name of the state.

    enabled
        Ensures whether or not VMotion should be enabled on a host as a boolean
        value where ``True`` indicates that VMotion should be enabled and ``False``
        indicates that VMotion should be disabled.

    device
        The device that uniquely identifies the VirtualNic that will be used for
        VMotion for the host. Defaults to ``vmk0``.

    Example:

    .. code-block:: yaml

        configure-vmotion:
          esxi.vmotion_configured:
            - enabled: True
            - device: sample-device

    """
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    esxi_cmd = "esxi.cmd"
    host = __pillar__["proxy"]["host"]

    current_vmotion_enabled = __salt__[esxi_cmd]("get_vmotion_enabled").get(host)
    current_vmotion_enabled = current_vmotion_enabled.get("VMotion Enabled")

    # Configure VMotion Enabled state, if changed.
    if enabled != current_vmotion_enabled:
        # Only run the command if not using test=True
        if not __opts__["test"]:
            # Enable VMotion if enabled=True
            if enabled is True:
                response = __salt__[esxi_cmd]("vmotion_enable", device=device).get(host)
                error = response.get("Error")
                if error:
                    ret["comment"] = f"Error: {error}"
                    return ret
            # Disable VMotion if enabled=False
            else:
                response = __salt__[esxi_cmd]("vmotion_disable").get(host)
                error = response.get("Error")
                if error:
                    ret["comment"] = f"Error: {error}"
                    return ret
        ret["changes"].update(
            {"enabled": {"old": current_vmotion_enabled, "new": enabled}}
        )

    ret["result"] = True
    if ret["changes"] == {}:
        ret["comment"] = "VMotion configuration is already in the desired state."
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "VMotion configuration will change."

    return ret


@_deprecation_message
def vsan_configured(name, enabled, add_disks_to_vsan=False):
    """
    Configures a host's VSAN properties such as enabling or disabling VSAN, or
    adding VSAN-eligible disks to the VSAN system for the host.

    name
        Name of the state.

    enabled
        Ensures whether or not VSAN should be enabled on a host as a boolean
        value where ``True`` indicates that VSAN should be enabled and ``False``
        indicates that VSAN should be disabled.

    add_disks_to_vsan
        If set to ``True``, any VSAN-eligible disks for the given host will be added
        to the host's VSAN system. Default is ``False``.

    Example:

    .. code-block:: yaml

        configure-host-vsan:
          esxi.vsan_configured:
            - enabled: True
            - add_disks_to_vsan: True

    """
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    esxi_cmd = "esxi.cmd"
    host = __pillar__["proxy"]["host"]

    current_vsan_enabled = __salt__[esxi_cmd]("get_vsan_enabled").get(host)
    error = current_vsan_enabled.get("Error")
    if error:
        ret["comment"] = f"Error: {error}"
        return ret
    current_vsan_enabled = current_vsan_enabled.get("VSAN Enabled")

    # Configure VSAN Enabled state, if changed.
    if enabled != current_vsan_enabled:
        # Only run the command if not using test=True
        if not __opts__["test"]:
            # Enable VSAN if enabled=True
            if enabled is True:
                response = __salt__[esxi_cmd]("vsan_enable").get(host)
                error = response.get("Error")
                if error:
                    ret["comment"] = f"Error: {error}"
                    return ret
            # Disable VSAN if enabled=False
            else:
                response = __salt__[esxi_cmd]("vsan_disable").get(host)
                error = response.get("Error")
                if error:
                    ret["comment"] = f"Error: {error}"
                    return ret
        ret["changes"].update(
            {"enabled": {"old": current_vsan_enabled, "new": enabled}}
        )

    # Add any eligible disks to VSAN, if requested.
    if add_disks_to_vsan:
        current_eligible_disks = __salt__[esxi_cmd]("get_vsan_eligible_disks").get(host)
        error = current_eligible_disks.get("Error")
        if error:
            ret["comment"] = f"Error: {error}"
            return ret

        disks = current_eligible_disks.get("Eligible")
        if disks and isinstance(disks, list):
            # Only run the command if not using test=True
            if not __opts__["test"]:
                response = __salt__[esxi_cmd]("vsan_add_disks").get(host)
                error = response.get("Error")
                if error:
                    ret["comment"] = f"Error: {error}"
                    return ret

            ret["changes"].update({"add_disks_to_vsan": {"old": "", "new": disks}})

    ret["result"] = True
    if ret["changes"] == {}:
        ret["comment"] = "VSAN configuration is already in the desired state."
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "VSAN configuration will change."

    return ret


@_deprecation_message
def ssh_configured(
    name,
    service_running,
    ssh_key=None,
    ssh_key_file=None,
    service_policy=None,
    service_restart=False,
    certificate_verify=None,
):
    """
    Manage the SSH configuration for a host including whether or not SSH is running or
    the presence of a given SSH key. Note: Only one ssh key can be uploaded for root.
    Uploading a second key will replace any existing key.

    name
        Name of the state.

    service_running
        Ensures whether or not the SSH service should be running on a host. Represented
        as a boolean value where ``True`` indicates that SSH should be running and
        ``False`` indicates that SSH should stopped.

        In order to update SSH keys, the SSH service must be running.

    ssh_key
        Public SSH key to added to the authorized_keys file on the ESXi host. You can
        use ``ssh_key`` or ``ssh_key_file``, but not both.

    ssh_key_file
        File containing the public SSH key to be added to the authorized_keys file on
        the ESXi host. You can use ``ssh_key_file`` or ``ssh_key``, but not both.

    service_policy
        The policy to set for the NTP service.

        .. note::

            When setting the service policy to ``off`` or ``on``, you *must* quote the
            setting. If you don't, the yaml parser will set the string to a boolean,
            which will cause trouble checking for stateful changes and will error when
            trying to set the policy on the ESXi host.

    service_restart
        If set to ``True``, the SSH service will be restarted, regardless of its
        previous running state. Default is ``False``.

    certificate_verify
        If set to ``True``, the SSL connection must present a valid certificate.
        Default is ``True``.

    Example:

    .. code-block:: yaml

        configure-host-ssh:
          esxi.ssh_configured:
            - service_running: True
            - ssh_key_file: /etc/salt/ssh_keys/my_key.pub
            - service_policy: 'on'
            - service_restart: True
            - certificate_verify: True

    """
    if certificate_verify is None:
        certificate_verify = True
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    esxi_cmd = "esxi.cmd"
    host = __pillar__["proxy"]["host"]
    ssh = "ssh"

    ssh_running = __salt__[esxi_cmd]("get_service_running", service_name=ssh).get(host)
    error = ssh_running.get("Error")
    if error:
        ret["comment"] = f"Error: {error}"
        return ret
    ssh_running = ssh_running.get(ssh)

    # Configure SSH service_running state, if changed.
    if service_running != ssh_running:
        # Only actually run the command if not using test=True
        if not __opts__["test"]:
            # Start SSH if service_running=True
            if service_running is True:
                enable = __salt__[esxi_cmd]("service_start", service_name=ssh).get(host)
                error = enable.get("Error")
                if error:
                    ret["comment"] = f"Error: {error}"
                    return ret
            # Disable SSH if service_running=False
            else:
                disable = __salt__[esxi_cmd]("service_stop", service_name=ssh).get(host)
                error = disable.get("Error")
                if error:
                    ret["comment"] = f"Error: {error}"
                    return ret

        ret["changes"].update(
            {"service_running": {"old": ssh_running, "new": service_running}}
        )

    # If uploading an SSH key or SSH key file, see if there's a current
    # SSH key and compare the current key to the key set in the state.
    current_ssh_key, ssh_key_changed = None, False
    if ssh_key or ssh_key_file:
        current_ssh_key = __salt__[esxi_cmd](
            "get_ssh_key", certificate_verify=certificate_verify
        )
        error = current_ssh_key.get("Error")
        if error:
            ret["comment"] = f"Error: {error}"
            return ret
        current_ssh_key = current_ssh_key.get("key")
        if current_ssh_key:
            clean_current_key = _strip_key(current_ssh_key).split(" ")
            if not ssh_key:
                ssh_key = ""
                # Open ssh key file and read in contents to create one key string
                with salt.utils.files.fopen(ssh_key_file, "r") as key_file:
                    for line in key_file:
                        if line.startswith("#"):
                            # Commented line
                            continue
                        ssh_key = ssh_key + line

            clean_ssh_key = _strip_key(ssh_key).split(" ")
            # Check that the first two list items of clean key lists are equal.
            if (
                clean_current_key[0] != clean_ssh_key[0]
                or clean_current_key[1] != clean_ssh_key[1]
            ):
                ssh_key_changed = True
        else:
            # If current_ssh_key is None, but we're setting a new key with
            # either ssh_key or ssh_key_file, then we need to flag the change.
            ssh_key_changed = True

    # Upload SSH key, if changed.
    if ssh_key_changed:
        if not __opts__["test"]:
            # Upload key
            response = __salt__[esxi_cmd](
                "upload_ssh_key",
                ssh_key=ssh_key,
                ssh_key_file=ssh_key_file,
                certificate_verify=certificate_verify,
            )
            error = response.get("Error")
            if error:
                ret["comment"] = f"Error: {error}"
                return ret
        ret["changes"].update(
            {
                "SSH Key": {
                    "old": current_ssh_key,
                    "new": ssh_key if ssh_key else ssh_key_file,
                }
            }
        )

    # Configure service_policy
    if service_policy:
        current_service_policy = __salt__[esxi_cmd](
            "get_service_policy", service_name=ssh
        ).get(host)
        error = current_service_policy.get("Error")
        if error:
            ret["comment"] = f"Error: {error}"
            return ret
        current_service_policy = current_service_policy.get(ssh)

        if service_policy != current_service_policy:
            # Only run the command if not using test=True
            if not __opts__["test"]:
                response = __salt__[esxi_cmd](
                    "set_service_policy",
                    service_name=ssh,
                    service_policy=service_policy,
                ).get(host)
                error = response.get("Error")
                if error:
                    ret["comment"] = f"Error: {error}"
                    return ret
            ret["changes"].update(
                {
                    "service_policy": {
                        "old": current_service_policy,
                        "new": service_policy,
                    }
                }
            )

    # Restart ssh_service if service_restart=True
    if service_restart:
        # Only run the command if not using test=True
        if not __opts__["test"]:
            response = __salt__[esxi_cmd]("service_restart", service_name=ssh).get(host)
            error = response.get("Error")
            if error:
                ret["comment"] = f"Error: {error}"
                return ret
        ret["changes"].update(
            {"service_restart": {"old": "", "new": "SSH service restarted."}}
        )

    ret["result"] = True
    if ret["changes"] == {}:
        ret["comment"] = "SSH service is already in the desired state."
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "SSH service state will change."

    return ret


@_deprecation_message
def syslog_configured(
    name,
    syslog_configs,
    firewall=True,
    reset_service=True,
    reset_syslog_config=False,
    reset_configs=None,
):
    """
    Ensures the specified syslog configuration parameters. By default,
    this state will reset the syslog service after any new or changed
    parameters are set successfully.

    name
        Name of the state.

    syslog_configs
        Name of parameter to set (corresponds to the command line switch for
        esxcli without the double dashes (--))

        Valid syslog_config values are ``logdir``, ``loghost``, ``logdir-unique``,
        ``default-rotate``, ``default-size``, and ``default-timeout``.

        Each syslog_config option also needs a configuration value to set.
        For example, ``loghost`` requires URLs or IP addresses to use for
        logging. Multiple log servers can be specified by listing them,
        comma-separated, but without spaces before or after commas

        (reference: https://blogs.vmware.com/vsphere/2012/04/configuring-multiple-syslog-servers-for-esxi-5.html)

    firewall
        Enable the firewall rule set for syslog. Defaults to ``True``.

    reset_service
        After a successful parameter set, reset the service. Defaults to ``True``.

    reset_syslog_config
        Resets the syslog service to its default settings. Defaults to ``False``.
        If set to ``True``, default settings defined by the list of syslog configs
        in ``reset_configs`` will be reset before running any other syslog settings.

    reset_configs
        A comma-delimited list of parameters to reset. Only runs if
        ``reset_syslog_config`` is set to ``True``. If ``reset_syslog_config`` is set
        to ``True``, but no syslog configs are listed in ``reset_configs``, then
        ``reset_configs`` will be set to ``all`` by default.

        See ``syslog_configs`` parameter above for a list of valid options.

    Example:

    .. code-block:: yaml

        configure-host-syslog:
          esxi.syslog_configured:
            - syslog_configs:
                loghost: ssl://localhost:5432,tcp://10.1.0.1:1514
                default-timeout: 120
            - firewall: True
            - reset_service: True
            - reset_syslog_config: True
            - reset_configs: loghost,default-timeout
    """
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    esxi_cmd = "esxi.cmd"
    host = __pillar__["proxy"]["host"]

    if reset_syslog_config:
        if not reset_configs:
            reset_configs = "all"
        # Only run the command if not using test=True
        if not __opts__["test"]:
            reset = __salt__[esxi_cmd](
                "reset_syslog_config", syslog_config=reset_configs
            ).get(host)
            for key, val in reset.items():
                if isinstance(val, bool):
                    continue
                if not val.get("success"):
                    msg = val.get("message")
                    if not msg:
                        msg = (
                            "There was an error resetting a syslog config '{}'."
                            "Please check debug logs.".format(val)
                        )
                    ret["comment"] = f"Error: {msg}"
                    return ret

        ret["changes"].update(
            {"reset_syslog_config": {"old": "", "new": reset_configs}}
        )

    current_firewall = __salt__[esxi_cmd]("get_firewall_status").get(host)
    error = current_firewall.get("Error")
    if error:
        ret["comment"] = f"Error: {error}"
        return ret

    current_firewall = current_firewall.get("rulesets").get("syslog")
    if current_firewall != firewall:
        # Only run the command if not using test=True
        if not __opts__["test"]:
            enabled = __salt__[esxi_cmd](
                "enable_firewall_ruleset",
                ruleset_enable=firewall,
                ruleset_name="syslog",
            ).get(host)
            if enabled.get("retcode") != 0:
                err = enabled.get("stderr")
                out = enabled.get("stdout")
                ret["comment"] = f"Error: {err if err else out}"
                return ret

        ret["changes"].update({"firewall": {"old": current_firewall, "new": firewall}})

    current_syslog_config = __salt__[esxi_cmd]("get_syslog_config").get(host)
    for key, val in syslog_configs.items():
        # The output of get_syslog_config has different keys than the keys
        # Used to set syslog_config values. We need to look them up first.
        try:
            lookup_key = _lookup_syslog_config(key)
        except KeyError:
            ret["comment"] = f"'{key}' is not a valid config variable."
            return ret

        current_val = current_syslog_config[lookup_key]
        if str(current_val) != str(val):
            # Only run the command if not using test=True
            if not __opts__["test"]:
                response = __salt__[esxi_cmd](
                    "set_syslog_config",
                    syslog_config=key,
                    config_value=val,
                    firewall=firewall,
                    reset_service=reset_service,
                ).get(host)
                success = response.get(key).get("success")
                if not success:
                    msg = response.get(key).get("message")
                    if not msg:
                        msg = (
                            "There was an error setting syslog config '{}'. "
                            "Please check debug logs.".format(key)
                        )
                    ret["comment"] = msg
                    return ret

            if not ret["changes"].get("syslog_config"):
                ret["changes"].update({"syslog_config": {}})
            ret["changes"]["syslog_config"].update(
                {key: {"old": current_val, "new": val}}
            )

    ret["result"] = True
    if ret["changes"] == {}:
        ret["comment"] = "Syslog is already in the desired state."
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Syslog state will change."

    return ret


@depends(HAS_PYVMOMI)
@depends(HAS_JSONSCHEMA)
@_deprecation_message
def diskgroups_configured(name, diskgroups, erase_disks=False):
    """
    Configures the disk groups to use for vsan.

    This function will do the following:

    1. Check whether or not all disks in the diskgroup spec exist, and raises
       and errors if they do not.

    2. Create diskgroups with the correct disk configurations if diskgroup
       (identified by the cache disk canonical name) doesn't exist

    3. Adds extra capacity disks to the existing diskgroup

    Example:

    .. code:: python

        {
            'cache_scsi_addr': 'vmhba1:C0:T0:L0',
            'capacity_scsi_addrs': [
                'vmhba2:C0:T0:L0',
                'vmhba3:C0:T0:L0',
                'vmhba4:C0:T0:L0',
            ]
        }

    name
        Mandatory state name

    diskgroups
        Disk group representation containing scsi disk addresses.
        Scsi addresses are expected for disks in the diskgroup:

    erase_disks
        Specifies whether to erase all partitions on all disks member of the
        disk group before the disk group is created. Default value is False.
    """
    proxy_details = __salt__["esxi.get_details"]()
    hostname = (
        proxy_details["host"]
        if not proxy_details.get("vcenter")
        else proxy_details["esxi_host"]
    )
    log.info("Running state %s for host '%s'", name, hostname)
    # Variable used to return the result of the invocation
    ret = {"name": name, "result": None, "changes": {}, "comments": None}
    # Signals if errors have been encountered
    errors = False
    # Signals if changes are required
    changes = False
    comments = []
    diskgroup_changes = {}
    si = None
    try:
        log.trace("Validating diskgroups_configured input")
        schema = DiskGroupsDiskScsiAddressSchema.serialize()
        try:
            jsonschema.validate(
                {"diskgroups": diskgroups, "erase_disks": erase_disks}, schema
            )
        except jsonschema.exceptions.ValidationError as exc:
            raise InvalidConfigError(exc)
        si = __salt__["vsphere.get_service_instance_via_proxy"]()
        host_disks = __salt__["vsphere.list_disks"](service_instance=si)
        if not host_disks:
            raise VMwareObjectRetrievalError(
                f"No disks retrieved from host '{hostname}'"
            )
        scsi_addr_to_disk_map = {d["scsi_address"]: d for d in host_disks}
        log.trace("scsi_addr_to_disk_map = %s", scsi_addr_to_disk_map)
        existing_diskgroups = __salt__["vsphere.list_diskgroups"](service_instance=si)
        cache_disk_to_existing_diskgroup_map = {
            dg["cache_disk"]: dg for dg in existing_diskgroups
        }
    except CommandExecutionError as err:
        log.error("Error: %s", err)
        if si:
            __salt__["vsphere.disconnect"](si)
        ret.update(
            {"result": False if not __opts__["test"] else None, "comment": str(err)}
        )
        return ret

    # Iterate through all of the disk groups
    for idx, dg in enumerate(diskgroups):
        # Check for cache disk
        if not dg["cache_scsi_addr"] in scsi_addr_to_disk_map:
            comments.append(
                "No cache disk with scsi address '{}' was found.".format(
                    dg["cache_scsi_addr"]
                )
            )
            log.error(comments[-1])
            errors = True
            continue

        # Check for capacity disks
        cache_disk_id = scsi_addr_to_disk_map[dg["cache_scsi_addr"]]["id"]
        cache_disk_display = "{} (id:{})".format(dg["cache_scsi_addr"], cache_disk_id)
        bad_scsi_addrs = []
        capacity_disk_ids = []
        capacity_disk_displays = []
        for scsi_addr in dg["capacity_scsi_addrs"]:
            if scsi_addr not in scsi_addr_to_disk_map:
                bad_scsi_addrs.append(scsi_addr)
                continue
            capacity_disk_ids.append(scsi_addr_to_disk_map[scsi_addr]["id"])
            capacity_disk_displays.append(f"{scsi_addr} (id:{capacity_disk_ids[-1]})")
        if bad_scsi_addrs:
            comments.append(
                "Error in diskgroup #{}: capacity disks with scsi addresses {} "
                "were not found.".format(
                    idx, ", ".join([f"'{a}'" for a in bad_scsi_addrs])
                )
            )
            log.error(comments[-1])
            errors = True
            continue

        if not cache_disk_to_existing_diskgroup_map.get(cache_disk_id):
            # A new diskgroup needs to be created
            log.trace("erase_disks = %s", erase_disks)
            if erase_disks:
                if __opts__["test"]:
                    comments.append(
                        "State {} will "
                        "erase all disks of disk group #{}; "
                        "cache disk: '{}', "
                        "capacity disk(s): {}."
                        "".format(
                            name,
                            idx,
                            cache_disk_display,
                            ", ".join([f"'{a}'" for a in capacity_disk_displays]),
                        )
                    )
                else:
                    # Erase disk group disks
                    for disk_id in [cache_disk_id] + capacity_disk_ids:
                        __salt__["vsphere.erase_disk_partitions"](
                            disk_id=disk_id, service_instance=si
                        )
                    comments.append(
                        "Erased disks of diskgroup #{}; "
                        "cache disk: '{}', capacity disk(s): "
                        "{}".format(
                            idx,
                            cache_disk_display,
                            ", ".join([f"'{a}'" for a in capacity_disk_displays]),
                        )
                    )
                    log.info(comments[-1])

            if __opts__["test"]:
                comments.append(
                    "State {} will create "
                    "the disk group #{}; cache disk: '{}', "
                    "capacity disk(s): {}.".format(
                        name,
                        idx,
                        cache_disk_display,
                        ", ".join([f"'{a}'" for a in capacity_disk_displays]),
                    )
                )
                log.info(comments[-1])
                changes = True
                continue
            try:
                __salt__["vsphere.create_diskgroup"](
                    cache_disk_id,
                    capacity_disk_ids,
                    safety_checks=False,
                    service_instance=si,
                )
            except VMwareSaltError as err:
                comments.append(f"Error creating disk group #{idx}: {err}.")
                log.error(comments[-1])
                errors = True
                continue

            comments.append(f"Created disk group #'{idx}'.")
            log.info(comments[-1])
            diskgroup_changes[str(idx)] = {
                "new": {"cache": cache_disk_display, "capacity": capacity_disk_displays}
            }
            changes = True
            continue

        # The diskgroup exists; checking the capacity disks
        log.debug(
            "Disk group #%s exists. Checking capacity disks: %s.",
            idx,
            capacity_disk_displays,
        )
        existing_diskgroup = cache_disk_to_existing_diskgroup_map.get(cache_disk_id)
        existing_capacity_disk_displays = [
            "{} (id:{})".format(
                [d["scsi_address"] for d in host_disks if d["id"] == disk_id][0],
                disk_id,
            )
            for disk_id in existing_diskgroup["capacity_disks"]
        ]
        # Populate added disks and removed disks and their displays
        added_capacity_disk_ids = []
        added_capacity_disk_displays = []
        removed_capacity_disk_ids = []
        removed_capacity_disk_displays = []
        for disk_id in capacity_disk_ids:
            if disk_id not in existing_diskgroup["capacity_disks"]:
                disk_scsi_addr = [
                    d["scsi_address"] for d in host_disks if d["id"] == disk_id
                ][0]
                added_capacity_disk_ids.append(disk_id)
                added_capacity_disk_displays.append(f"{disk_scsi_addr} (id:{disk_id})")
        for disk_id in existing_diskgroup["capacity_disks"]:
            if disk_id not in capacity_disk_ids:
                disk_scsi_addr = [
                    d["scsi_address"] for d in host_disks if d["id"] == disk_id
                ][0]
                removed_capacity_disk_ids.append(disk_id)
                removed_capacity_disk_displays.append(
                    f"{disk_scsi_addr} (id:{disk_id})"
                )

        log.debug(
            "Disk group #%s: existing capacity disk ids: %s; added "
            "capacity disk ids: %s; removed capacity disk ids: %s",
            idx,
            existing_capacity_disk_displays,
            added_capacity_disk_displays,
            removed_capacity_disk_displays,
        )

        # TODO revisit this when removing capacity disks is supported
        if removed_capacity_disk_ids:
            comments.append(
                "Error removing capacity disk(s) {} from disk group #{}; "
                "operation is not supported."
                "".format(
                    ", ".join([f"'{id}'" for id in removed_capacity_disk_displays]),
                    idx,
                )
            )
            log.error(comments[-1])
            errors = True
            continue

        if added_capacity_disk_ids:
            # Capacity disks need to be added to disk group

            # Building a string representation of the capacity disks
            # that need to be added
            s = ", ".join([f"'{id}'" for id in added_capacity_disk_displays])
            if __opts__["test"]:
                comments.append(
                    "State {} will add capacity disk(s) {} to disk group #{}.".format(
                        name, s, idx
                    )
                )
                log.info(comments[-1])
                changes = True
                continue
            try:
                __salt__["vsphere.add_capacity_to_diskgroup"](
                    cache_disk_id,
                    added_capacity_disk_ids,
                    safety_checks=False,
                    service_instance=si,
                )
            except VMwareSaltError as err:
                comments.append(
                    "Error adding capacity disk(s) {} to disk group #{}: {}.".format(
                        s, idx, err
                    )
                )
                log.error(comments[-1])
                errors = True
                continue

            com = f"Added capacity disk(s) {s} to disk group #{idx}"
            log.info(com)
            comments.append(com)
            diskgroup_changes[str(idx)] = {
                "new": {
                    "cache": cache_disk_display,
                    "capacity": capacity_disk_displays,
                },
                "old": {
                    "cache": cache_disk_display,
                    "capacity": existing_capacity_disk_displays,
                },
            }
            changes = True
            continue

        # No capacity needs to be added
        s = f"Disk group #{idx} is correctly configured. Nothing to be done."
        log.info(s)
        comments.append(s)
    __salt__["vsphere.disconnect"](si)

    # Build the final return message
    result = (
        True
        if not (changes or errors)
        else (
            None  # no changes/errors
            if __opts__["test"]
            else False if errors else True  # running in test mode
        )
    )  # found errors; defaults to True
    ret.update(
        {"result": result, "comment": "\n".join(comments), "changes": diskgroup_changes}
    )
    return ret


@depends(HAS_PYVMOMI)
@depends(HAS_JSONSCHEMA)
@_deprecation_message
def host_cache_configured(
    name,
    enabled,
    datastore,
    swap_size="100%",
    dedicated_backing_disk=False,
    erase_backing_disk=False,
):
    """
    Configures the host cache used for swapping.

    It will do the following:

    1. Checks if backing disk exists

    2. Creates the VMFS datastore if doesn't exist (datastore partition will be
       created and use the entire disk)

    3. Raises an error if ``dedicated_backing_disk`` is ``True`` and partitions
       already exist on the backing disk

    4. Configures host_cache to use a portion of the datastore for caching
       (either a specific size or a percentage of the datastore)

    Examples

    Percentage swap size (can't be 100%)

    .. code:: python

        {
            'enabled': true,
            'datastore': {
                'backing_disk_scsi_addr': 'vmhba0:C0:T0:L0',
                'vmfs_version': 5,
                'name': 'hostcache'
                }
            'dedicated_backing_disk': false
            'swap_size': '98%',
        }

    Fixed sized swap size

    .. code:: python

        {
            'enabled': true,
            'datastore': {
                'backing_disk_scsi_addr': 'vmhba0:C0:T0:L0',
                'vmfs_version': 5,
                'name': 'hostcache'
                }
            'dedicated_backing_disk': true
            'swap_size': '10GiB',
        }

    name
        Mandatory state name.

    enabled
        Specifies whether the host cache is enabled.

    datastore
        Specifies the host cache datastore.

    swap_size
        Specifies the size of the host cache swap. Can be a percentage or a
        value in GiB. Default value is ``100%``.

    dedicated_backing_disk
        Specifies whether the backing disk is dedicated to the host cache which
        means it must have no other partitions. Default is False

    erase_backing_disk
        Specifies whether to erase all partitions on the backing disk before
        the datastore is created. Default value is False.
    """
    log.trace("enabled = %s", enabled)
    log.trace("datastore = %s", datastore)
    log.trace("swap_size = %s", swap_size)
    log.trace("erase_backing_disk = %s", erase_backing_disk)
    # Variable used to return the result of the invocation
    proxy_details = __salt__["esxi.get_details"]()
    hostname = (
        proxy_details["host"]
        if not proxy_details.get("vcenter")
        else proxy_details["esxi_host"]
    )
    log.trace("hostname = %s", hostname)
    log.info("Running host_cache_swap_configured for host '%s'", hostname)
    ret = {
        "name": hostname,
        "comment": "Default comments",
        "result": None,
        "changes": {},
    }
    result = None if __opts__["test"] else True  # We assume success
    needs_setting = False
    comments = []
    changes = {}
    si = None
    try:
        log.debug("Validating host_cache_configured input")
        schema = HostCacheSchema.serialize()
        try:
            jsonschema.validate(
                {
                    "enabled": enabled,
                    "datastore": datastore,
                    "swap_size": swap_size,
                    "erase_backing_disk": erase_backing_disk,
                },
                schema,
            )
        except jsonschema.exceptions.ValidationError as exc:
            raise InvalidConfigError(exc)
        m = re.match(r"(\d+)(%|GiB)", swap_size)
        swap_size_value = int(m.group(1))
        swap_type = m.group(2)
        log.trace("swap_size_value = %s; swap_type = %s", swap_size_value, swap_type)
        si = __salt__["vsphere.get_service_instance_via_proxy"]()
        host_cache = __salt__["vsphere.get_host_cache"](service_instance=si)

        # Check enabled
        if host_cache["enabled"] != enabled:
            changes.update({"enabled": {"old": host_cache["enabled"], "new": enabled}})
            needs_setting = True

        # Check datastores
        existing_datastores = None
        if host_cache.get("datastore"):
            existing_datastores = __salt__["vsphere.list_datastores_via_proxy"](
                datastore_names=[datastore["name"]], service_instance=si
            )
        # Retrieve backing disks
        existing_disks = __salt__["vsphere.list_disks"](
            scsi_addresses=[datastore["backing_disk_scsi_addr"]], service_instance=si
        )
        if not existing_disks:
            raise VMwareObjectRetrievalError(
                "Disk with scsi address '{}' was not found in host '{}'".format(
                    datastore["backing_disk_scsi_addr"], hostname
                )
            )
        backing_disk = existing_disks[0]
        backing_disk_display = "{} (id:{})".format(
            backing_disk["scsi_address"], backing_disk["id"]
        )
        log.trace("backing_disk = %s", backing_disk_display)

        existing_datastore = None
        if not existing_datastores:
            # Check if disk needs to be erased
            if erase_backing_disk:
                if __opts__["test"]:
                    comments.append(
                        "State {} will erase the backing disk '{}' on host '{}'.".format(
                            name, backing_disk_display, hostname
                        )
                    )
                    log.info(comments[-1])
                else:
                    # Erase disk
                    __salt__["vsphere.erase_disk_partitions"](
                        disk_id=backing_disk["id"], service_instance=si
                    )
                    comments.append(
                        "Erased backing disk '{}' on host '{}'.".format(
                            backing_disk_display, hostname
                        )
                    )
                    log.info(comments[-1])
            # Create the datastore
            if __opts__["test"]:
                comments.append(
                    "State {} will create the datastore '{}', with backing disk "
                    "'{}', on host '{}'.".format(
                        name, datastore["name"], backing_disk_display, hostname
                    )
                )
                log.info(comments[-1])
            else:
                if dedicated_backing_disk:
                    # Check backing disk doesn't already have partitions
                    partitions = __salt__["vsphere.list_disk_partitions"](
                        disk_id=backing_disk["id"], service_instance=si
                    )
                    log.trace("partitions = %s", partitions)
                    # We will ignore the mbr partitions
                    non_mbr_partitions = [p for p in partitions if p["format"] != "mbr"]
                    if len(non_mbr_partitions) > 0:
                        raise VMwareApiError(
                            "Backing disk '{}' has unexpected partitions".format(
                                backing_disk_display
                            )
                        )
                __salt__["vsphere.create_vmfs_datastore"](
                    datastore["name"],
                    existing_disks[0]["id"],
                    datastore["vmfs_version"],
                    service_instance=si,
                )
                comments.append(
                    "Created vmfs datastore '{}', backed by "
                    "disk '{}', on host '{}'.".format(
                        datastore["name"], backing_disk_display, hostname
                    )
                )
                log.info(comments[-1])
                changes.update(
                    {
                        "datastore": {
                            "new": {
                                "name": datastore["name"],
                                "backing_disk": backing_disk_display,
                            }
                        }
                    }
                )
                existing_datastore = __salt__["vsphere.list_datastores_via_proxy"](
                    datastore_names=[datastore["name"]], service_instance=si
                )[0]
            needs_setting = True
        else:
            # Check datastore is backed by the correct disk
            if not existing_datastores[0].get("backing_disk_ids"):
                raise VMwareSaltError(
                    "Datastore '{}' doesn't have a backing disk".format(
                        datastore["name"]
                    )
                )
            if backing_disk["id"] not in existing_datastores[0]["backing_disk_ids"]:

                raise VMwareSaltError(
                    "Datastore '{}' is not backed by the correct disk: "
                    "expected '{}'; got {}".format(
                        datastore["name"],
                        backing_disk["id"],
                        ", ".join(
                            [
                                f"'{disk}'"
                                for disk in existing_datastores[0]["backing_disk_ids"]
                            ]
                        ),
                    )
                )

            comments.append(
                "Datastore '{}' already exists on host '{}' "
                "and is backed by disk '{}'. Nothing to be "
                "done.".format(datastore["name"], hostname, backing_disk_display)
            )
            existing_datastore = existing_datastores[0]
            log.trace("existing_datastore = %s", existing_datastore)
            log.info(comments[-1])

        if existing_datastore:
            # The following comparisons can be done if the existing_datastore
            # is set; it may not be set if running in test mode
            #
            # We support percent, as well as MiB, we will convert the size
            # to MiB, multiples of 1024 (VMware SDK limitation)
            if swap_type == "%":
                # Percentage swap size
                # Convert from bytes to MiB
                raw_size_MiB = (swap_size_value / 100.0) * (
                    existing_datastore["capacity"] / 1024 / 1024
                )
            else:
                raw_size_MiB = swap_size_value * 1024
            log.trace("raw_size = %sMiB", raw_size_MiB)
            swap_size_MiB = int(raw_size_MiB / 1024) * 1024
            log.trace("adjusted swap_size = %sMiB", swap_size_MiB)
            existing_swap_size_MiB = 0
            m = (
                re.match(r"(\d+)MiB", host_cache.get("swap_size"))
                if host_cache.get("swap_size")
                else None
            )
            if m:
                # if swap_size from the host is set and has an expected value
                # we are going to parse it to get the number of MiBs
                existing_swap_size_MiB = int(m.group(1))
            if not existing_swap_size_MiB == swap_size_MiB:
                needs_setting = True
                changes.update(
                    {
                        "swap_size": {
                            "old": f"{existing_swap_size_MiB / 1024}GiB",
                            "new": f"{swap_size_MiB / 1024}GiB",
                        }
                    }
                )

        if needs_setting:
            if __opts__["test"]:
                comments.append(
                    "State {} will configure the host cache on host '{}' to: {}.".format(
                        name,
                        hostname,
                        {
                            "enabled": enabled,
                            "datastore_name": datastore["name"],
                            "swap_size": swap_size,
                        },
                    )
                )
            else:
                if (existing_datastore["capacity"] / 1024.0**2) < swap_size_MiB:

                    raise ArgumentValueError(
                        "Capacity of host cache datastore '{}' ({} MiB) is "
                        "smaller than the required swap size ({} MiB)".format(
                            existing_datastore["name"],
                            existing_datastore["capacity"] / 1024.0**2,
                            swap_size_MiB,
                        )
                    )
                __salt__["vsphere.configure_host_cache"](
                    enabled,
                    datastore["name"],
                    swap_size_MiB=swap_size_MiB,
                    service_instance=si,
                )
                comments.append(f"Host cache configured on host '{hostname}'.")
        else:
            comments.append(
                "Host cache on host '{}' is already correctly "
                "configured. Nothing to be done.".format(hostname)
            )
            result = True
        __salt__["vsphere.disconnect"](si)
        log.info(comments[-1])
        ret.update(
            {"comment": "\n".join(comments), "result": result, "changes": changes}
        )
        return ret
    except CommandExecutionError as err:
        log.error("Error: %s.", err)
        if si:
            __salt__["vsphere.disconnect"](si)
        ret.update(
            {
                "result": False if not __opts__["test"] else None,
                "comment": f"{err}.",
            }
        )
        return ret


def _lookup_syslog_config(config):
    """
    Helper function that looks up syslog_config keys available from
    ``vsphere.get_syslog_config``.
    """
    lookup = {
        "default-timeout": "Default Network Retry Timeout",
        "logdir": "Local Log Output",
        "default-size": "Local Logging Default Rotation Size",
        "logdir-unique": "Log To Unique Subdirectory",
        "default-rotate": "Local Logging Default Rotations",
        "loghost": "Remote Host",
    }

    return lookup.get(config)


def _strip_key(key_string):
    """
    Strips an SSH key string of white space and line endings and returns the new string.

    key_string
        The string to be stripped.
    """
    key_string.strip()
    key_string.replace("\n", "")
    key_string.replace("\r\n", "")
    return key_string



import copy


def iter_transport_opts(opts):
    """
    Yield transport, opts for all master configured transports
    """
    transports = set()

    for transport, opts_overrides in opts.get("transport_opts", {}).items():
        t_opts = copy.deepcopy(opts)
        t_opts.update(opts_overrides)
        t_opts["transport"] = transport
        transports.add(transport)
        yield transport, t_opts

    if opts["transport"] not in transports:
        yield opts["transport"], opts



"""
State module to manage Elasticsearch indices

.. versionadded:: 2015.8.0
.. deprecated:: 2017.7.0
   Use elasticsearch state instead
"""

import logging

log = logging.getLogger(__name__)


def absent(name):
    """
    Ensure that the named index is absent.

    name
        Name of the index to remove
    """

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    try:
        index = __salt__["elasticsearch.index_get"](index=name)
        if index and name in index:
            if __opts__["test"]:
                ret["comment"] = f"Index {name} will be removed"
                ret["changes"]["old"] = index[name]
                ret["result"] = None
            else:
                ret["result"] = __salt__["elasticsearch.index_delete"](index=name)
                if ret["result"]:
                    ret["comment"] = f"Successfully removed index {name}"
                    ret["changes"]["old"] = index[name]
                else:
                    ret["comment"] = (
                        f"Failed to remove index {name} for unknown reasons"
                    )
        else:
            ret["comment"] = f"Index {name} is already absent"
    except Exception as err:  # pylint: disable=broad-except
        ret["result"] = False
        ret["comment"] = str(err)

    return ret


def present(name, definition=None):
    """
    .. versionadded:: 2015.8.0
    .. versionchanged:: 2017.3.0
        Marked ``definition`` as optional.

    Ensure that the named index is present.

    name
        Name of the index to add

    definition
        Optional dict for creation parameters as per https://www.elastic.co/guide/en/elasticsearch/reference/current/indices-create-index.html

    **Example:**

    .. code-block:: yaml

        # Default settings
        mytestindex:
          elasticsearch_index.present

        # Extra settings
        mytestindex2:
          elasticsearch_index.present:
            - definition:
                settings:
                  index:
                    number_of_shards: 10
    """

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    try:
        index_exists = __salt__["elasticsearch.index_exists"](index=name)
        if not index_exists:
            if __opts__["test"]:
                ret["comment"] = "Index {} does not exist and will be created".format(
                    name
                )
                ret["changes"] = {"new": definition}
                ret["result"] = None
            else:
                output = __salt__["elasticsearch.index_create"](
                    index=name, body=definition
                )
                if output:
                    ret["comment"] = f"Successfully created index {name}"
                    ret["changes"] = {
                        "new": __salt__["elasticsearch.index_get"](index=name)[name]
                    }
                else:
                    ret["result"] = False
                    ret["comment"] = f"Cannot create index {name}, {output}"
        else:
            ret["comment"] = f"Index {name} is already present"
    except Exception as err:  # pylint: disable=broad-except
        ret["result"] = False
        ret["comment"] = str(err)

    return ret



"""
Management of ipsets
======================

This is an ipset-specific module designed to manage IPSets for use
in IPTables Firewalls.

.. code-block:: yaml

    setname:
      ipset.set_present:
        - set_type: bitmap:ip
        - range: 192.168.0.0/16
        - comment: True

    setname:
      ipset.set_absent:
        - set_type: bitmap:ip
        - range: 192.168.0.0/16
        - comment: True

    setname_entries:
      ipset.present:
        - set_name: setname
        - entry: 192.168.0.3
        - comment: Hello
        - require:
            - ipset: baz

    setname_entries:
      ipset.present:
        - set_name: setname
        - entry:
            - 192.168.0.3
            - 192.168.1.3
        - comment: Hello
        - require:
            - ipset: baz

    setname_entries:
      ipset.absent:
        - set_name: setname
        - entry:
            - 192.168.0.3
            - 192.168.1.3
        - comment: Hello
        - require:
            - ipset: baz

    setname:
      ipset.flush:

"""

import logging

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if the ipset module is available in __salt__
    """
    if "ipset.version" in __salt__:
        return True
    return (False, "ipset module could not be loaded")


def set_present(name, set_type, family="ipv4", **kwargs):
    """
    .. versionadded:: 2014.7.0

    Verify the set exists.

    name
        A user-defined set name.

    set_type
        The type for the set.

    family
        Networking family, either ipv4 or ipv6
    """

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    set_check = __salt__["ipset.check_set"](name)
    if set_check is True:
        ret["result"] = True
        ret["comment"] = f"ipset set {name} already exists for {family}"
        return ret

    if __opts__["test"]:
        ret["comment"] = f"ipset set {name} would be added for {family}"
        return ret
    command = __salt__["ipset.new_set"](name, set_type, family, **kwargs)
    if command is True:
        ret["changes"] = {"locale": name}
        ret["result"] = True
        ret["comment"] = f"ipset set {name} created successfully for {family}"
        return ret
    else:
        ret["result"] = False
        ret["comment"] = "Failed to create set {0} for {2}: {1}".format(
            name, command.strip(), family
        )
        return ret


def set_absent(name, family="ipv4", **kwargs):
    """
    .. versionadded:: 2014.7.0

    Verify the set is absent.

    family
        Networking family, either ipv4 or ipv6
    """

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    set_check = __salt__["ipset.check_set"](name, family)
    if not set_check:
        ret["result"] = True
        ret["comment"] = f"ipset set {name} for {family} is already absent"
        return ret
    if __opts__["test"]:
        ret["comment"] = f"ipset set {name} for {family} would be removed"
        return ret
    flush_set = __salt__["ipset.flush"](name, family)
    if flush_set:
        command = __salt__["ipset.delete_set"](name, family)
        if command is True:
            ret["changes"] = {"locale": name}
            ret["result"] = True
            ret["comment"] = "ipset set {} deleted successfully for family {}".format(
                name, family
            )
        else:
            ret["result"] = False
            ret["comment"] = "Failed to delete set {0} for {2}: {1}".format(
                name, command.strip(), family
            )
    else:
        ret["result"] = False
        ret["comment"] = "Failed to flush set {0} for {2}: {1}".format(
            name, flush_set.strip(), family
        )
    return ret


def present(name, entry=None, family="ipv4", **kwargs):
    """
    .. versionadded:: 2014.7.0

    Append a entry to a set

    name
        A user-defined name to call this entry by in another part of a state or
        formula. This should not be an actual entry.

    entry
        A single entry to add to a set or a list of entries to add to a set

    family
        Network family, ipv4 or ipv6.

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if not entry:
        ret["result"] = False
        ret["comment"] = "ipset entry must be specified"
        return ret

    entries = []
    if isinstance(entry, list):
        entries = entry
    else:
        entries.append(entry)

    for entry in entries:
        entry_opts = ""
        if " " in entry:
            entry, entry_opts = entry.split(" ", 1)
        if "timeout" in kwargs and "timeout" not in entry_opts:
            entry_opts = "timeout {} {}".format(kwargs["timeout"], entry_opts)
        if "comment" in kwargs and "comment" not in entry_opts:
            entry_opts = "{} comment {}".format(entry_opts, kwargs["comment"])
        _entry = " ".join([entry, entry_opts.lstrip()]).strip()

        if __salt__["ipset.check"](kwargs["set_name"], _entry, family) is True:
            ret["comment"] += "entry for {} already in set {} for {}\n".format(
                entry, kwargs["set_name"], family
            )
        else:
            if __opts__["test"]:
                ret["result"] = None
                ret[
                    "comment"
                ] += "entry {} would be added to set {} for family {}\n".format(
                    entry, kwargs["set_name"], family
                )
            else:
                command = __salt__["ipset.add"](
                    kwargs["set_name"], _entry, family, **kwargs
                )
                if "Error" not in command:
                    ret["changes"] = {"locale": name}
                    ret["comment"] += "entry {} added to set {} for family {}\n".format(
                        _entry, kwargs["set_name"], family
                    )
                else:
                    ret["result"] = False
                    ret["comment"] = (
                        "Failed to add to entry {1} to set {0} for family {2}.\n{3}".format(
                            kwargs["set_name"], _entry, family, command
                        )
                    )
    return ret


def absent(name, entry=None, entries=None, family="ipv4", **kwargs):
    """
    .. versionadded:: 2014.7.0

    Remove a entry or entries from a chain

    name
        A user-defined name to call this entry by in another part of a state or
        formula. This should not be an actual entry.

    family
        Network family, ipv4 or ipv6.

    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if not entry:
        ret["result"] = False
        ret["comment"] = "ipset entry must be specified"
        return ret

    entries = []
    if isinstance(entry, list):
        entries = entry
    else:
        entries.append(entry)

    for entry in entries:
        entry_opts = ""
        if " " in entry:
            entry, entry_opts = entry.split(" ", 1)
        if "timeout" in kwargs and "timeout" not in entry_opts:
            entry_opts = "timeout {} {}".format(kwargs["timeout"], entry_opts)
        if "comment" in kwargs and "comment" not in entry_opts:
            entry_opts = "{} comment {}".format(entry_opts, kwargs["comment"])
        _entry = " ".join([entry, entry_opts]).strip()

        log.debug("_entry %s", _entry)
        if not __salt__["ipset.check"](kwargs["set_name"], _entry, family) is True:
            ret["result"] = True
            ret[
                "comment"
            ] += "ipset entry for {} not present in set {} for {}\n".format(
                _entry, kwargs["set_name"], family
            )
        else:
            if __opts__["test"]:
                ret["result"] = None
                ret[
                    "comment"
                ] += "ipset entry {} would be removed from set {} for {}\n".format(
                    entry, kwargs["set_name"], family
                )
            else:
                command = __salt__["ipset.delete"](
                    kwargs["set_name"], entry, family, **kwargs
                )
                if "Error" not in command:
                    ret["changes"] = {"locale": name}
                    ret["result"] = True
                    ret[
                        "comment"
                    ] += "ipset entry {} removed from set {} for {}\n".format(
                        _entry, kwargs["set_name"], family
                    )
                else:
                    ret["result"] = False
                    ret["comment"] = (
                        "Failed to delete ipset entry from set {} for {}. "
                        "Attempted entry was {}.\n"
                        "{}\n".format(kwargs["set_name"], family, _entry, command)
                    )
    return ret


def flush(name, family="ipv4", **kwargs):
    """
    .. versionadded:: 2014.7.0

    Flush current ipset set

    family
        Networking family, either ipv4 or ipv6

    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    set_check = __salt__["ipset.check_set"](name)
    if set_check is False:
        ret["result"] = False
        ret["comment"] = f"ipset set {name} does not exist for {family}"
        return ret

    if __opts__["test"]:
        ret["comment"] = "ipset entries in set {} for {} would be flushed".format(
            name, family
        )
        return ret
    if __salt__["ipset.flush"](name, family):
        ret["changes"] = {"locale": name}
        ret["result"] = True
        ret["comment"] = f"Flushed ipset entries from set {name} for {family}"
        return ret
    else:
        ret["result"] = False
        ret["comment"] = "Failed to flush ipset entries from set {} for {}".format(
            name, family
        )
        return ret



"""
On-demand logging
=================

.. versionadded:: 2017.7.0

The sole purpose of this module is logging messages in the (proxy) minion.
It comes very handy when debugging complex Jinja templates, for example:

.. code-block:: jinja

    {%- for var in range(10) %}
      {%- do salt["log.info"](var) -%}
    {%- endfor %}

CLI Example:

.. code-block:: bash

    salt '*' log.error "Please don't do that, this module is not for CLI use!"
"""

import logging

log = logging.getLogger(__name__)

__virtualname__ = "log"
__proxyenabled__ = ["*"]


def __virtual__():
    return __virtualname__


def debug(message):
    """
    Log message at level DEBUG.
    """
    log.debug(message)
    return True


def info(message):
    """
    Log message at level INFO.
    """
    log.info(message)
    return True


def warning(message):
    """
    Log message at level WARNING.
    """
    log.warning(message)
    return True


def error(message):
    """
    Log message at level ERROR.
    """
    log.error(message)
    return True


def critical(message):
    """
    Log message at level CRITICAL.
    """
    log.critical(message)
    return True


def exception(message):
    """
    Log message at level EXCEPTION.
    """
    log.exception(message)
    return True



"""
Test utility methods that communicate with SMB shares.
"""

import getpass
import logging
import os
import shutil
import signal
import subprocess
import tempfile
import time

import pytest

import salt.utils.files
import salt.utils.network
import salt.utils.path
import salt.utils.smb
from tests.support.case import TestCase

log = logging.getLogger(__name__)
CONFIG = (
    "[global]\n"
    "realm = saltstack.com\n"
    "interfaces = lo 127.0.0.0/8\n"
    "smb ports = 1445\n"
    "log level = 2\n"
    "map to guest = Bad User\n"
    "enable core files = no\n"
    "passdb backend = smbpasswd\n"
    "smb passwd file = {passwdb}\n"
    "lock directory = {samba_dir}\n"
    "state directory = {samba_dir}\n"
    "cache directory = {samba_dir}\n"
    "pid directory = {samba_dir}\n"
    "private dir = {samba_dir}\n"
    "ncalrpc dir = {samba_dir}\n"
    "socket options = IPTOS_LOWDELAY TCP_NODELAY\n"
    "min receivefile size = 0\n"
    "write cache size = 0\n"
    "client ntlmv2 auth = no\n"
    "client min protocol = SMB3_11\n"
    "client plaintext auth = no\n"
    "\n"
    "[public]\n"
    "path = {public_dir}\n"
    "read only = no\n"
    "guest ok = no\n"
    "writeable = yes\n"
    "force user = {user}\n"
)
TBE = (
    "{}:0:XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX:AC8E657F8"
    "3DF82BEEA5D43BDAF7800CC:[U          ]:LCT-507C14C7:"
)
IPV6_ENABLED = bool(salt.utils.network.ip_addrs6(include_loopback=True))


@pytest.mark.skipif(
    not salt.utils.smb.HAS_SMBPROTOCOL, reason='"smbprotocol" needs to be installed.'
)
@pytest.mark.skip_if_binaries_missing("smbd")
class TestSmb(TestCase):

    _smbd = None

    @staticmethod
    def check_pid(pid):
        try:
            os.kill(pid, 0)
        except OSError:
            return False
        else:
            return True

    @classmethod
    def setUpClass(cls):
        tmpdir = tempfile.mkdtemp()
        cls.samba_dir = os.path.join(tmpdir, "samba")
        cls.public_dir = os.path.join(tmpdir, "public")
        os.makedirs(cls.samba_dir)
        os.makedirs(cls.public_dir)
        os.chmod(cls.samba_dir, 0o775)
        os.chmod(cls.public_dir, 0o775)
        passwdb = os.path.join(tmpdir, "passwdb")
        cls.username = getpass.getuser()
        with salt.utils.files.fopen(passwdb, "w") as fp:
            fp.write(TBE.format(cls.username))
        samba_conf = os.path.join(tmpdir, "smb.conf")
        with salt.utils.files.fopen(samba_conf, "w") as fp:
            fp.write(
                CONFIG.format(
                    samba_dir=cls.samba_dir,
                    public_dir=cls.public_dir,
                    passwdb=passwdb,
                    user=cls.username,
                )
            )
        cls._smbd = subprocess.Popen(
            [shutil.which("smbd"), "-FS", "-P0", "-s", samba_conf]
        )
        time.sleep(1)
        pidfile = os.path.join(cls.samba_dir, "smbd.pid")
        with salt.utils.files.fopen(pidfile, "r") as fp:
            cls._pid = int(fp.read().strip())
        if not cls.check_pid(cls._pid):
            raise Exception("Unable to locate smbd's pid file")

    @classmethod
    def tearDownClass(cls):
        log.warning("teardown")
        os.kill(cls._pid, signal.SIGTERM)

    def test_write_file_ipv4(self):
        """
        Transfer a file over SMB
        """
        name = "test_write_file_v4.txt"
        content = "write test file content ipv4"
        share_path = os.path.join(self.public_dir, name)
        assert not os.path.exists(share_path)

        local_path = tempfile.mktemp()
        with salt.utils.files.fopen(local_path, "w") as fp:
            fp.write(content)
        conn = salt.utils.smb.get_conn("127.0.0.1", self.username, "foo", port=1445)
        salt.utils.smb.put_file(local_path, name, "public", conn=conn)
        conn.close()

        assert os.path.exists(share_path)
        with salt.utils.files.fopen(share_path, "r") as fp:
            result = fp.read()
        assert result == content

    @pytest.mark.skipif(not IPV6_ENABLED, reason="IPv6 not enabled")
    def test_write_file_ipv6(self):
        """
        Transfer a file over SMB
        """
        name = "test_write_file_v6.txt"
        content = "write test file content ipv6"
        share_path = os.path.join(self.public_dir, name)
        assert not os.path.exists(share_path)

        local_path = tempfile.mktemp()
        with salt.utils.files.fopen(local_path, "w") as fp:
            fp.write(content)
        conn = salt.utils.smb.get_conn("::1", self.username, "foo", port=1445)
        salt.utils.smb.put_file(local_path, name, "public", conn=conn)
        conn.close()

        assert os.path.exists(share_path)
        with salt.utils.files.fopen(share_path, "r") as fp:
            result = fp.read()
        assert result == content

    def test_write_str_v4(self):
        """
        Write a string to a file over SMB
        """
        name = "test_write_str.txt"
        content = "write test file content"
        share_path = os.path.join(self.public_dir, name)
        assert not os.path.exists(share_path)
        conn = salt.utils.smb.get_conn("127.0.0.1", self.username, "foo", port=1445)
        salt.utils.smb.put_str(content, name, "public", conn=conn)
        conn.close()

        assert os.path.exists(share_path)
        with salt.utils.files.fopen(share_path, "r") as fp:
            result = fp.read()
        assert result == content

    @pytest.mark.skipif(not IPV6_ENABLED, reason="IPv6 not enabled")
    def test_write_str_v6(self):
        """
        Write a string to a file over SMB
        """
        name = "test_write_str_v6.txt"
        content = "write test file content"
        share_path = os.path.join(self.public_dir, name)
        assert not os.path.exists(share_path)
        conn = salt.utils.smb.get_conn("::1", self.username, "foo", port=1445)
        salt.utils.smb.put_str(content, name, "public", conn=conn)
        conn.close()

        assert os.path.exists(share_path)
        with salt.utils.files.fopen(share_path, "r") as fp:
            result = fp.read()
        assert result == content

    def test_delete_file_v4(self):
        """
        Validate deletion of files over SMB
        """
        name = "test_delete_file.txt"
        content = "read test file content"
        share_path = os.path.join(self.public_dir, name)
        with salt.utils.files.fopen(share_path, "w") as fp:
            fp.write(content)
        assert os.path.exists(share_path)

        conn = salt.utils.smb.get_conn("127.0.0.1", self.username, "foo", port=1445)
        salt.utils.smb.delete_file(name, "public", conn=conn)
        conn.close()

        assert not os.path.exists(share_path)

    @pytest.mark.skipif(not IPV6_ENABLED, reason="IPv6 not enabled")
    def test_delete_file_v6(self):
        """
        Validate deletion of files over SMB
        """
        name = "test_delete_file_v6.txt"
        content = "read test file content"
        share_path = os.path.join(self.public_dir, name)
        with salt.utils.files.fopen(share_path, "w") as fp:
            fp.write(content)
        assert os.path.exists(share_path)

        conn = salt.utils.smb.get_conn("::1", self.username, "foo", port=1445)
        salt.utils.smb.delete_file(name, "public", conn=conn)
        conn.close()

        assert not os.path.exists(share_path)

    def test_mkdirs_v4(self):
        """
        Create directories over SMB
        """
        dir_name = "mkdirs/test"
        share_path = os.path.join(self.public_dir, dir_name)
        assert not os.path.exists(share_path)

        conn = salt.utils.smb.get_conn("127.0.0.1", self.username, "foo", port=1445)
        salt.utils.smb.mkdirs(dir_name, "public", conn=conn)
        conn.close()

        assert os.path.exists(share_path)

    @pytest.mark.skipif(not IPV6_ENABLED, reason="IPv6 not enabled")
    def test_mkdirs_v6(self):
        """
        Create directories over SMB
        """
        dir_name = "mkdirs/testv6"
        share_path = os.path.join(self.public_dir, dir_name)
        assert not os.path.exists(share_path)

        conn = salt.utils.smb.get_conn("::1", self.username, "foo", port=1445)
        salt.utils.smb.mkdirs(dir_name, "public", conn=conn)
        conn.close()

        assert os.path.exists(share_path)

    def test_delete_dirs_v4(self):
        """
        Validate deletion of directoreies over SMB
        """
        dir_name = "deldirs"
        subdir_name = "deldirs/test"
        local_path = os.path.join(self.public_dir, subdir_name)
        os.makedirs(local_path)
        assert os.path.exists(local_path)

        conn = salt.utils.smb.get_conn("127.0.0.1", self.username, "foo", port=1445)
        salt.utils.smb.delete_directory(subdir_name, "public", conn=conn)
        conn.close()

        conn = salt.utils.smb.get_conn("127.0.0.1", self.username, "foo", port=1445)
        salt.utils.smb.delete_directory(dir_name, "public", conn=conn)
        conn.close()

        assert not os.path.exists(local_path)
        assert not os.path.exists(os.path.join(self.public_dir, dir_name))

    @pytest.mark.skipif(not IPV6_ENABLED, reason="IPv6 not enabled")
    def test_delete_dirs_v6(self):
        """
        Validate deletion of directoreies over SMB
        """
        dir_name = "deldirsv6"
        subdir_name = "deldirsv6/test"
        local_path = os.path.join(self.public_dir, subdir_name)
        os.makedirs(local_path)
        assert os.path.exists(local_path)

        conn = salt.utils.smb.get_conn("::1", self.username, "foo", port=1445)
        salt.utils.smb.delete_directory(subdir_name, "public", conn=conn)
        conn.close()

        conn = salt.utils.smb.get_conn("::1", self.username, "foo", port=1445)
        salt.utils.smb.delete_directory(dir_name, "public", conn=conn)
        conn.close()

        assert not os.path.exists(local_path)
        assert not os.path.exists(os.path.join(self.public_dir, dir_name))

    def test_connection(self):
        """
        Validate creation of an SMB connection
        """
        conn = salt.utils.smb.get_conn("127.0.0.1", self.username, "foo", port=1445)
        conn.close()

    @pytest.mark.skipif(not IPV6_ENABLED, reason="IPv6 not enabled")
    def test_connection_v6(self):
        """
        Validate creation of an SMB connection
        """
        conn = salt.utils.smb.get_conn("::1", self.username, "foo", port=1445)
        conn.close()



"""
    :codeauthor: :email:`Anthony Shaw <anthonyshaw@apache.org>`
"""

import pytest

import salt.proxy.napalm as napalm_proxy
import tests.support.napalm as napalm_test_support
from tests.support.mock import MagicMock, patch


@pytest.fixture
def test_opts():
    return {
        "proxytype": "napalm",
        "driver": "junos",
        "host": "core05.nrt02",
        "id": "core05.nrt02",
    }


def mock_get_device(opts, *args, **kwargs):
    return {"DRIVER": napalm_test_support.MockNapalmDevice(), "UP": True}


@pytest.fixture
def configure_loader_modules(test_opts):
    module_globals = {
        "__salt__": {
            "config.option": MagicMock(
                return_value={"test": {"driver": "test", "key": "2orgk34kgk34g"}}
            ),
            "__context__": {},
        }
    }
    module_globals["napalm_base"] = MagicMock()
    with patch("salt.utils.napalm.get_device", mock_get_device):
        yield {napalm_proxy: module_globals}


def test_init(test_opts):
    ret = napalm_proxy.init(test_opts)
    assert ret is True


def test_alive(test_opts):
    napalm_proxy.init(test_opts)
    ret = napalm_proxy.alive(test_opts)
    assert ret is True


def test_ping(test_opts):

    # UP is True
    mock_context = {
        "napalm_device": {
            "opts": {
                "proxytype": "napalm",
                "driver": "junos",
                "host": "core05.nrt02",
                "id": "core05.nrt02",
            },
            "id": "core05.nrt02",
            "network_device": {"DRIVER": "", "UP": True},
            "details": {"initialized": True},
        }
    }

    with patch.dict(napalm_proxy.__context__, mock_context):
        ret = napalm_proxy.ping()
        assert ret is True

    # UP is False
    mock_context = {
        "napalm_device": {
            "opts": {
                "proxytype": "napalm",
                "driver": "junos",
                "host": "core05.nrt02",
                "id": "core05.nrt02",
            },
            "id": "core05.nrt02",
            "network_device": {"DRIVER": "", "UP": False},
            "details": {"initialized": True},
        }
    }

    with patch.dict(napalm_proxy.__context__, mock_context):
        ret = napalm_proxy.ping()
        assert not ret

    # UP is missing
    mock_context = {
        "napalm_device": {
            "opts": {
                "proxytype": "napalm",
                "driver": "junos",
                "host": "core05.nrt02",
                "id": "core05.nrt02",
            },
            "id": "core05.nrt02",
            "network_device": {"DRIVER": ""},
            "details": {"initialized": True},
        }
    }

    with patch.dict(napalm_proxy.__context__, mock_context):
        ret = napalm_proxy.ping()
        assert not ret

    # network_device is missing
    mock_context = {
        "napalm_device": {
            "opts": {
                "proxytype": "napalm",
                "driver": "junos",
                "host": "core05.nrt02",
                "id": "core05.nrt02",
            },
            "id": "core05.nrt02",
            "details": {"initialized": True},
        }
    }

    with patch.dict(napalm_proxy.__context__, mock_context):
        ret = napalm_proxy.ping()
        assert not ret

    # napalm_device is missing
    mock_context = {}

    with patch.dict(napalm_proxy.__context__, mock_context):
        ret = napalm_proxy.ping()
        assert not ret


def test_initialized(test_opts):
    # initialized is True
    mock_context = {
        "napalm_device": {
            "opts": {
                "proxytype": "napalm",
                "driver": "junos",
                "host": "core05.nrt02",
                "id": "core05.nrt02",
            },
            "id": "core05.nrt02",
            "network_device": {"DRIVER": "", "UP": True},
            "details": {"initialized": True},
        }
    }

    with patch.dict(napalm_proxy.__context__, mock_context):
        ret = napalm_proxy.initialized()
        assert ret is True

    # initialized is False
    mock_context = {
        "napalm_device": {
            "opts": {
                "proxytype": "napalm",
                "driver": "junos",
                "host": "core05.nrt02",
                "id": "core05.nrt02",
            },
            "id": "core05.nrt02",
            "network_device": {"DRIVER": "", "UP": False},
            "details": {"initialized": False},
        }
    }

    with patch.dict(napalm_proxy.__context__, mock_context):
        ret = napalm_proxy.initialized()
        assert not ret

    # initialized is missing
    mock_context = {
        "napalm_device": {
            "opts": {
                "proxytype": "napalm",
                "driver": "junos",
                "host": "core05.nrt02",
                "id": "core05.nrt02",
            },
            "id": "core05.nrt02",
            "network_device": {"DRIVER": ""},
            "details": {},
        }
    }

    with patch.dict(napalm_proxy.__context__, mock_context):
        ret = napalm_proxy.initialized()
        assert not ret

    # details is missing
    mock_context = {
        "napalm_device": {
            "opts": {
                "proxytype": "napalm",
                "driver": "junos",
                "host": "core05.nrt02",
                "id": "core05.nrt02",
            },
            "id": "core05.nrt02",
        }
    }

    with patch.dict(napalm_proxy.__context__, mock_context):
        ret = napalm_proxy.initialized()
        assert not ret

    # napalm_device is missing
    mock_context = {}

    with patch.dict(napalm_proxy.__context__, mock_context):
        ret = napalm_proxy.initialized()
        assert not ret


def test_get_device(test_opts):
    # network_devie exists
    mock_context = {
        "napalm_device": {
            "opts": {
                "proxytype": "napalm",
                "driver": "junos",
                "host": "core05.nrt02",
                "id": "core05.nrt02",
            },
            "id": "core05.nrt02",
            "network_device": {"DRIVER": "", "UP": True},
            "details": {"initialized": True},
        }
    }

    with patch.dict(napalm_proxy.__context__, mock_context):
        ret = napalm_proxy.get_device()
        assert ret == {"DRIVER": "", "UP": True}

    # network_device is missing
    mock_context = {
        "napalm_device": {
            "opts": {
                "proxytype": "napalm",
                "driver": "junos",
                "host": "core05.nrt02",
                "id": "core05.nrt02",
            },
            "id": "core05.nrt02",
            "details": {"initialized": True},
        }
    }

    with patch.dict(napalm_proxy.__context__, mock_context):
        ret = napalm_proxy.get_device()
        assert not ret

    # napalm_device is missing
    mock_context = {}

    with patch.dict(napalm_proxy.__context__, mock_context):
        ret = napalm_proxy.get_device()
        assert not ret


def test_get_grains(test_opts):
    napalm_proxy.init(test_opts)
    ret = napalm_proxy.get_grains()
    assert ret["out"] == napalm_test_support.TEST_FACTS.copy()


def test_grains_refresh(test_opts):
    napalm_proxy.init(test_opts)
    ret = napalm_proxy.grains_refresh()
    assert ret["out"] == napalm_test_support.TEST_FACTS.copy()


def test_fns():
    ret = napalm_proxy.fns()
    assert "details" in ret


def test_shutdown(test_opts):
    napalm_proxy.init(test_opts)
    ret = napalm_proxy.shutdown(test_opts)
    assert ret is True


def test_call(test_opts):
    # network_devie exists
    mock_context = {
        "napalm_device": {
            "opts": {
                "proxytype": "napalm",
                "driver": "junos",
                "host": "core05.nrt02",
                "id": "core05.nrt02",
            },
            "id": "core05.nrt02",
            "network_device": mock_get_device(test_opts),
            "details": {"initialized": True},
        }
    }

    with patch.dict(napalm_proxy.__context__, mock_context):
        ret = napalm_proxy.call("get_arp_table")
        assert ret["out"] == napalm_test_support.TEST_ARP_TABLE.copy()

    # network_device is missing
    mock_context = {}
    with patch.dict(napalm_proxy.__context__, mock_context):
        ret = napalm_proxy.call("get_arp_table")
        assert ret == {"result": False, "comment": "Not initialised yet", "out": None}



"""
Management of PostgreSQL schemas
================================

The postgres_schemas module is used to create and manage Postgres schemas.

.. code-block:: yaml

    public:
      postgres_schema.present 'dbname' 'name'
"""

import logging

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if the postgres module is present
    """
    if "postgres.schema_exists" not in __salt__:
        return (
            False,
            "Unable to load postgres module.  Make sure `postgres.bins_dir` is set.",
        )
    return True


def present(
    dbname,
    name,
    owner=None,
    user=None,
    db_user=None,
    db_password=None,
    db_host=None,
    db_port=None,
):
    """
    Ensure that the named schema is present in the database.

    dbname
        The database's name will work on

    name
        The name of the schema to manage

    user
        system user all operations should be performed on behalf of

    db_user
        database username if different from config or default

    db_password
        user password if any password for a specified user

    db_host
        Database host if different from config or default

    db_port
        Database port if different from config or default
    """
    ret = {
        "dbname": dbname,
        "name": name,
        "changes": {},
        "result": True,
        "comment": f"Schema {name} is already present in database {dbname}",
    }

    db_args = {
        "db_user": db_user,
        "db_password": db_password,
        "db_host": db_host,
        "db_port": db_port,
        "user": user,
    }

    # check if schema exists
    schema_attr = __salt__["postgres.schema_get"](dbname, name, **db_args)

    cret = None

    # The schema is not present, make it!
    if schema_attr is None:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "Schema {} is set to be created in database {}.".format(
                name, dbname
            )
            return ret
        cret = __salt__["postgres.schema_create"](dbname, name, owner=owner, **db_args)
    else:
        msg = "Schema {0} already exists in database {1}"
        cret = None

    if cret:
        msg = "Schema {0} has been created in database {1}"
        ret["result"] = True
        ret["changes"][name] = "Present"
    elif cret is not None:
        msg = "Failed to create schema {0} in database {1}"
        ret["result"] = False
    else:
        msg = "Schema {0} already exists in database {1}"
        ret["result"] = True

    ret["comment"] = msg.format(name, dbname)
    return ret


def absent(
    dbname, name, user=None, db_user=None, db_password=None, db_host=None, db_port=None
):
    """
    Ensure that the named schema is absent.

    dbname
        The database's name will work on

    name
        The name of the schema to remove

    user
        system user all operations should be performed on behalf of

    db_user
        database username if different from config or default

    db_password
        user password if any password for a specified user

    db_host
        Database host if different from config or default

    db_port
        Database port if different from config or default
    """
    ret = {"name": name, "dbname": dbname, "changes": {}, "result": True, "comment": ""}

    db_args = {
        "db_user": db_user,
        "db_password": db_password,
        "db_host": db_host,
        "db_port": db_port,
        "user": user,
    }

    # check if schema exists and remove it
    if __salt__["postgres.schema_exists"](dbname, name, **db_args):
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "Schema {} is set to be removed from database {}".format(
                name, dbname
            )
            return ret
        elif __salt__["postgres.schema_remove"](dbname, name, **db_args):
            ret["comment"] = "Schema {} has been removed from database {}".format(
                name, dbname
            )
            ret["changes"][name] = "Absent"
            return ret
        else:
            ret["result"] = False
            ret["comment"] = f"Schema {name} failed to be removed"
            return ret
    else:
        ret["comment"] = (
            "Schema {} is not present in database {}, so it cannot be removed".format(
                name, dbname
            )
        )

    return ret



import io
from textwrap import dedent

import pytest

import salt.modules.solaris_shadow as solaris_shadow
from tests.support.mock import MagicMock, patch

try:
    import pwd

    missing_pwd = False
except ImportError:
    pwd = None
    missing_pwd = True

try:
    import spwd  # pylint: disable=unused-import

    missing_spwd = False
except ImportError:
    missing_spwd = True


skip_on_missing_spwd = pytest.mark.skipif(
    missing_spwd, reason="Has no spwd module for accessing /etc/shadow passwords"
)
skip_on_missing_pwd = pytest.mark.skipif(
    missing_pwd, reason="Has no pwd module for accessing /etc/password passwords"
)

# pylint: disable=singleton-comparison

# TODO: A lot of the shadow functionality is common across solaris and Linux.
# It would be possible to combine some of this into salt/utils -W. Werner, 2021-01-26


@pytest.fixture
def configure_loader_modules():
    return {solaris_shadow: {"pwd": pwd}}


@pytest.fixture
def fake_fopen_has_etc_shadow():
    contents = dedent(
        """\
            foo:bar:bang
            whatever:is:shadow
            roscivs:bottia:bloop
        """
    )
    fake_output_shadow_file = io.StringIO()

    def fopen(file, mode, *args, **kwargs):
        for line in contents.split():
            if "b" in mode:
                return io.BytesIO(contents.encode())
            elif "w" in mode:
                return fake_output_shadow_file
            else:
                return io.StringIO(contents)

    with patch("salt.utils.files.fopen", side_effect=fopen, autospec=True):
        with patch.object(fake_output_shadow_file, "close"):
            yield fake_output_shadow_file
            fake_output_shadow_file.close()


@pytest.fixture
def has_spwd():
    with patch.object(solaris_shadow, "HAS_SPWD", True):
        yield


@pytest.fixture
def has_not_spwd():
    with patch.object(solaris_shadow, "HAS_SPWD", False):
        yield


@pytest.fixture
def fake_spnam():
    with patch(
        "spwd.getspnam",
        autospec=True,
    ) as fake_spnam:
        yield fake_spnam


@pytest.fixture
def fake_pwnam():
    with patch(
        "pwd.getpwnam",
        autospec=True,
    ) as fake_pwnam:
        yield fake_pwnam


@pytest.fixture
def has_shadow_file():
    with patch("os.path.isfile", return_value=True):
        yield


@pytest.fixture
def has_not_shadow_file():
    with patch("os.path.isfile", return_value=False):
        yield


@skip_on_missing_spwd
def test_when_spwd_module_exists_results_should_be_returned_from_getspnam(
    has_spwd, fake_spnam
):
    expected_results = {
        "name": "roscivs",
        "passwd": "bottia",
        "lstchg": "2010-08-14",
        "min": 0,
        "max": 42,
        "warn": "nope",
        "inact": "whatever",
        "expire": "never!",
    }
    fake_spnam.return_value.sp_nam = expected_results["name"]
    fake_spnam.return_value.sp_pwd = expected_results["passwd"]
    fake_spnam.return_value.sp_lstchg = expected_results["lstchg"]
    fake_spnam.return_value.sp_min = expected_results["min"]
    fake_spnam.return_value.sp_max = expected_results["max"]
    fake_spnam.return_value.sp_warn = expected_results["warn"]
    fake_spnam.return_value.sp_inact = expected_results["inact"]
    fake_spnam.return_value.sp_expire = expected_results["expire"]

    actual_results = solaris_shadow.info(name="roscivs")

    assert actual_results == expected_results


@skip_on_missing_spwd
def test_when_swpd_module_exists_and_no_results_then_results_should_be_empty(
    has_spwd, fake_spnam
):
    expected_results = {
        "name": "",
        "passwd": "",
        "lstchg": "",
        "min": "",
        "max": "",
        "warn": "",
        "inact": "",
        "expire": "",
    }
    fake_spnam.side_effect = KeyError

    actual_results = solaris_shadow.info(name="roscivs")

    assert actual_results == expected_results


@skip_on_missing_pwd
def test_when_pwd_fallback_is_used_and_no_name_exists_results_should_be_empty(
    has_not_spwd, fake_pwnam
):
    expected_results = {
        "name": "",
        "passwd": "",
        "lstchg": "",
        "min": "",
        "max": "",
        "warn": "",
        "inact": "",
        "expire": "",
    }
    fake_pwnam.side_effect = KeyError

    actual_results = solaris_shadow.info(name="wayne")

    assert actual_results == expected_results


@skip_on_missing_pwd
def test_when_etc_shadow_does_not_exist_info_should_be_empty_except_for_name(
    has_not_spwd, fake_pwnam, has_not_shadow_file
):
    expected_results = {
        "name": "wayne",
        "passwd": "",
        "lstchg": "",
        "min": "",
        "max": "",
        "warn": "",
        "inact": "",
        "expire": "",
    }
    fake_pwnam.return_value.pw_name = "not this name"

    actual_results = solaris_shadow.info(name="wayne")

    assert actual_results == expected_results


@skip_on_missing_pwd
def test_when_etc_shadow_exists_but_name_not_in_shadow_passwd_field_should_be_empty(
    fake_fopen_has_etc_shadow, has_not_spwd, fake_pwnam, has_shadow_file
):
    with patch.dict(
        solaris_shadow.__salt__,
        {"cmd.run_all": MagicMock(return_value={"retcode": 42})},
    ):
        actual_result = solaris_shadow.info(name="badname")

    assert actual_result["passwd"] == ""


@skip_on_missing_pwd
def test_when_name_in_etc_shadow_passwd_should_be_in_info(
    fake_fopen_has_etc_shadow, has_not_spwd, fake_pwnam, has_shadow_file
):
    with patch.dict(
        solaris_shadow.__salt__,
        {"cmd.run_all": MagicMock(return_value={"retcode": 42})},
    ):
        actual_result = solaris_shadow.info(name="roscivs")

    assert actual_result["passwd"] == "bottia"


def test_when_set_password_and_not_has_shadow_ret_should_be_empty_dict(
    has_not_shadow_file,
):
    actual_result = solaris_shadow.set_password(name="fnord", password="blarp")

    assert actual_result == {}


def test_set_password_should_return_False_if_passwd_in_info_is_different_than_new_password(
    has_shadow_file, fake_fopen_has_etc_shadow
):
    existing_password = "Fnord"
    failed_set_password = "ignore me"
    with patch(
        "salt.modules.solaris_shadow.info",
        autospec=True,
        return_value={"passwd": existing_password},
    ):
        actual_result = solaris_shadow.set_password(
            name="roscivs", password=failed_set_password
        )

        assert actual_result == False


@skip_on_missing_spwd
def test_when_set_password_and_name_in_shadow_then_password_should_be_changed_for_that_user(
    has_shadow_file, fake_fopen_has_etc_shadow, has_spwd, fake_spnam
):
    expected_password = "bottia2"
    expected_shadow_contents = dedent(
        """\
            foo:bar:bang
            whatever:is:shadow
            roscivs:bottia2:bloop
        """
    )
    with patch(
        "salt.modules.solaris_shadow.info",
        autospec=True,
        return_value={"passwd": expected_password},
    ):
        actual_result = solaris_shadow.set_password(
            name="roscivs", password=expected_password
        )

    assert fake_fopen_has_etc_shadow.getvalue() == expected_shadow_contents
    assert actual_result == True



"""
Module for configuring DNS Client on Windows systems
"""

import logging

import salt.utils.platform

try:
    import wmi

    import salt.utils.winapi

    HAS_LIBS = True
except ImportError:
    HAS_LIBS = False

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only works on Windows systems
    """
    if not salt.utils.platform.is_windows():
        return False, "Module win_dns_client: module only works on Windows systems"
    if not HAS_LIBS:
        return False, "Module win_dns_client: missing required libraries"
    return "win_dns_client"


def get_dns_servers(interface="Local Area Connection"):
    """
    Return a list of the configured DNS servers of the specified interface

    Args:
        interface (str): The name of the network interface. This is the name as
        it appears in the Control Panel under Network Connections

    Returns:
        list: A list of dns servers

    CLI Example:

    .. code-block:: bash

        salt '*' win_dns_client.get_dns_servers 'Local Area Connection'
    """
    # remove any escape characters
    interface = interface.split("\\")
    interface = "".join(interface)

    with salt.utils.winapi.Com():
        c = wmi.WMI()
        for iface in c.Win32_NetworkAdapter(NetEnabled=True):
            if interface == iface.NetConnectionID:
                iface_config = c.Win32_NetworkAdapterConfiguration(
                    Index=iface.Index
                ).pop()
                try:
                    return list(iface_config.DNSServerSearchOrder)
                except TypeError:
                    return []
    log.debug('Interface "%s" not found', interface)
    return False


def rm_dns(ip, interface="Local Area Connection"):
    """
    Remove the DNS server from the network interface

    CLI Example:

    .. code-block:: bash

        salt '*' win_dns_client.rm_dns <ip> <interface>
    """
    cmd = [
        "netsh",
        "interface",
        "ip",
        "delete",
        "dns",
        f'"{interface}"',
        f'"{ip}"',
        "validate=no",
    ]
    return __salt__["cmd.retcode"](cmd, python_shell=False) == 0


def add_dns(ip, interface="Local Area Connection", index=1):
    """
    Add the DNS server to the network interface
    (index starts from 1)

    Note: if the interface DNS is configured by DHCP, all the DNS servers will
    be removed from the interface and the requested DNS will be the only one

    CLI Example:

    .. code-block:: bash

        salt '*' win_dns_client.add_dns <ip> <interface> <index>
    """
    servers = get_dns_servers(interface)

    # Return False if could not find the interface
    if servers is False:
        return False

    # Return true if configured
    try:
        if servers[index - 1] == ip:
            return True
    except IndexError:
        pass

    # If configured in the wrong order delete it
    if ip in servers:
        rm_dns(ip, interface)

    cmd = [
        "netsh",
        "interface",
        "ip",
        "add",
        "dns",
        f'"{interface}"',
        f'"{ip}"',
        f"index={index}",
        "validate=no",
    ]

    return __salt__["cmd.retcode"](cmd, python_shell=False) == 0


def dns_dhcp(interface="Local Area Connection"):
    """
    Configure the interface to get its DNS servers from the DHCP server

    CLI Example:

    .. code-block:: bash

        salt '*' win_dns_client.dns_dhcp <interface>
    """
    cmd = ["netsh", "interface", "ip", "set", "dns", f'"{interface}"', "source=dhcp"]
    return __salt__["cmd.retcode"](cmd, python_shell=False) == 0


def get_dns_config(interface="Local Area Connection"):
    """
    Get the type of DNS configuration (dhcp / static).

    Args:
        interface (str): The name of the network interface. This is the
        Description in the Network Connection Details for the device

    Returns:
        bool: ``True`` if DNS is configured, otherwise ``False``

    CLI Example:

    .. code-block:: bash

        salt '*' win_dns_client.get_dns_config 'Local Area Connection'
    """
    # remove any escape characters
    interface = interface.split("\\")
    interface = "".join(interface)

    with salt.utils.winapi.Com():
        c = wmi.WMI()
        for iface in c.Win32_NetworkAdapterConfiguration(IPEnabled=1):
            if interface == iface.Description:
                return iface.DHCPEnabled



"""
States for solrcloud alias and collection configuration

.. versionadded:: 2017.7.0

"""

import salt.utils.json


def alias(name, collections, **kwargs):
    """
    Create alias and enforce collection list.

    Use the solrcloud module to get alias members and set them.

    You can pass additional arguments that will be forwarded to http.query

    name
        The collection name
    collections
        list of collections to include in the alias
    """
    ret = {
        "name": name,
        "changes": {},
        "result": False,
        "comment": "",
    }

    if __salt__["solrcloud.alias_exists"](name, **kwargs):
        alias_content = __salt__["solrcloud.alias_get_collections"](name, **kwargs)
        diff = set(alias_content).difference(set(collections))

        if len(diff) == 0:
            ret["result"] = True
            ret["comment"] = "Alias is in desired state"
            return ret

        if __opts__["test"]:
            ret["comment"] = f'The alias "{name}" will be updated.'
            ret["result"] = None
        else:
            __salt__["solrcloud.alias_set_collections"](name, collections, **kwargs)
            ret["comment"] = f'The alias "{name}" has been updated.'
            ret["result"] = True

        ret["changes"] = {
            "old": ",".join(alias_content),
            "new": ",".join(collections),
        }

    else:
        if __opts__["test"]:
            ret["comment"] = f'The alias "{name}" will be created.'
            ret["result"] = None
        else:
            __salt__["solrcloud.alias_set_collections"](name, collections, **kwargs)
            ret["comment"] = f'The alias "{name}" has been created.'
            ret["result"] = True

        ret["changes"] = {
            "old": None,
            "new": ",".join(collections),
        }

    return ret


def collection(name, options=None, **kwargs):
    """
    Create collection and enforce options.

    Use the solrcloud module to get collection parameters.

    You can pass additional arguments that will be forwarded to http.query

    name
        The collection name
    options : {}
        options to ensure
    """
    ret = {
        "name": name,
        "changes": {},
        "result": False,
        "comment": "",
    }

    if options is None:
        options = {}

    if __salt__["solrcloud.collection_exists"](name, **kwargs):

        diff = {}

        current_options = __salt__["solrcloud.collection_get_options"](name, **kwargs)

        # Filter options that can be updated
        updatable_options = [
            "maxShardsPerNode",
            "replicationFactor",
            "autoAddReplicas",
            "collection.configName",
            "rule",
            "snitch",
        ]

        options = [k for k in options.items() if k in updatable_options]

        for key, value in options:
            if key not in current_options or current_options[key] != value:
                diff[key] = value

        if len(diff) == 0:
            ret["result"] = True
            ret["comment"] = "Collection options are in desired state"
            return ret

        else:

            if __opts__["test"]:
                ret["comment"] = f'Collection options "{name}" will be changed.'
                ret["result"] = None
            else:
                __salt__["solrcloud.collection_set_options"](name, diff, **kwargs)
                ret["comment"] = 'Parameters were updated for collection "{}".'.format(
                    name
                )
                ret["result"] = True

            ret["changes"] = {
                "old": salt.utils.json.dumps(
                    current_options, sort_keys=True, indent=4, separators=(",", ": ")
                ),
                "new": salt.utils.json.dumps(
                    options, sort_keys=True, indent=4, separators=(",", ": ")
                ),
            }
            return ret

    else:

        new_changes = salt.utils.json.dumps(
            options, sort_keys=True, indent=4, separators=(",", ": ")
        )
        if __opts__["test"]:
            ret["comment"] = f'The collection "{name}" will be created.'
            ret["result"] = None
        else:
            __salt__["solrcloud.collection_create"](name, options, **kwargs)
            ret["comment"] = f'The collection "{name}" has been created.'
            ret["result"] = True

        ret["changes"] = {
            "old": None,
            "new": "options=" + new_changes,
        }

    return ret



"""
Windows Object Access Control Lists

Ensure an ACL is present
    parameters:
        name - the path of the object
        objectType - Registry/File/Directory
        user - user account or SID for the ace
        permission - permission for the ace (see module win_acl for available permissions for each objectType)
        acetype -  Allow/Deny
        propagation - how the ACL should apply to child objects (see module win_acl for available propagation types)

    .. code-block:: yaml

        addAcl:
          win_dacl.present:
            - name: HKEY_LOCAL_MACHINE\\SOFTWARE\\mykey
            - objectType: Registry
            - user: FakeUser
            - permission: FullControl
            - acetype: ALLOW
            - propagation: KEY&SUBKEYS

Ensure an ACL does not exist
    parameters:
        name - the path of the object
        objectType - Registry/File/Directory
        user - user account or SID for the ace
        permission - permission for the ace (see module win_acl for available permissions for each objectType)
        acetype -  Allow/Deny
        propagation - how the ACL should apply to child objects (see module win_acl for available propagation types)

    .. code-block:: yaml

        removeAcl:
          win_dacl.absent:
            - name: HKEY_LOCAL_MACHINE\\SOFTWARE\\mykey
            - objectType: Registry
            - user: FakeUser
            - permission: FulLControl
            - acetype: ALLOW
            - propagation: KEY&SUBKEYS

Ensure an object is inheriting permissions
    parameters:
        name - the path of the object
        objectType - Registry/File/Directory
        clear_existing_acl - True/False - when inheritance is enabled, should the existing ACL be kept or cleared out

    .. code-block:: yaml

        eInherit:
          win_dacl.enableinheritance:
            - name: HKEY_LOCAL_MACHINE\\SOFTWARE\\mykey
            - objectType: Registry
            - clear_existing_acl: True

Ensure an object is not inheriting permissions
    parameters:
        name - the path of the object
        objectType - Registry/File/Directory
        copy_inherited_acl - True/False - if inheritance is enabled, should the inherited permissions be copied to the ACL when inheritance is disabled

    .. code-block:: yaml

        dInherit:
          win_dacl.disableinheritance:
            - name: HKEY_LOCAL_MACHINE\\SOFTWARE\\mykey
            - objectType: Registry
            - copy_inherited_acl: False
"""

__virtualname__ = "win_dacl"


def __virtual__():
    """
    Load this state if the win_acl module exists
    """
    if "win_dacl.add_ace" in __salt__:
        return "win_dacl"
    return (False, "win_dacl module could not be loaded")


def present(name, objectType, user, permission, acetype, propagation):
    """
    Ensure an ACE is present
    """
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}
    tRet = __salt__["win_dacl.check_ace"](
        name, objectType, user, permission, acetype, propagation, True
    )
    if tRet["result"]:
        if not tRet["Exists"]:
            if __opts__["test"]:
                ret["result"] = None
                ret["comment"] = "The ACE is set to be added."
                ret["changes"]["Added ACEs"] = "{} {} {} on {}".format(
                    user, acetype, permission, propagation
                )
                return ret
            addRet = __salt__["win_dacl.add_ace"](
                name, objectType, user, permission, acetype, propagation
            )
            if addRet["result"]:
                ret["result"] = True
                ret["changes"] = dict(ret["changes"], **addRet["changes"])
            else:
                ret["result"] = False
                ret["comment"] = " ".join([ret["comment"], addRet["comment"]])
        else:
            if __opts__["test"]:
                ret["result"] = None
                ret["comment"] = "The ACE is present."
    else:
        ret["result"] = False
        ret["comment"] = tRet["comment"]
    return ret


def absent(name, objectType, user, permission, acetype, propagation):
    """
    Ensure an ACL does not exist
    """
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}
    tRet = __salt__["win_dacl.check_ace"](
        name, objectType, user, permission, acetype, propagation, True
    )
    if tRet["result"]:
        if tRet["Exists"]:
            if __opts__["test"]:
                ret["result"] = None
                ret["comment"] = "The ACE is set to be removed."
                ret["changes"]["Removed ACEs"] = "{} {} {} on {}".format(
                    user, acetype, permission, propagation
                )
                return ret
            addRet = __salt__["win_dacl.rm_ace"](
                name, objectType, user, permission, acetype, propagation
            )
            if addRet["result"]:
                ret["result"] = True
                ret["changes"] = dict(ret["changes"], **addRet["changes"])
            else:
                ret["result"] = False
                ret["comment"] = " ".join([ret["comment"], addRet["comment"]])
        else:
            if __opts__["test"]:
                ret["result"] = None
                ret["comment"] = "The ACE is not present."
    else:
        ret["result"] = False
        ret["comment"] = tRet["comment"]
    return ret


def inherit(name, objectType, clear_existing_acl=False):
    """
    Ensure an object is inheriting ACLs from its parent
    """
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}
    tRet = __salt__["win_dacl.check_inheritance"](name, objectType)
    if tRet["result"]:
        if not tRet["Inheritance"]:
            if __opts__["test"]:
                ret["result"] = None
                ret["changes"]["Inheritance"] = "Enabled"
                ret["comment"] = "Inheritance is set to be enabled."
                ret["changes"]["Existing ACLs"] = (
                    "Are set to be removed"
                    if clear_existing_acl
                    else "Are set to be kept"
                )
                return ret
            eRet = __salt__["win_dacl.enable_inheritance"](
                name, objectType, clear_existing_acl
            )
            if eRet["result"]:
                ret["result"] = True
                ret["changes"] = dict(ret["changes"], **eRet["changes"])
            else:
                ret["result"] = False
                ret["comment"] = " ".join([ret["comment"], eRet["comment"]])
        else:
            if __opts__["test"]:
                ret["result"] = None
                ret["comment"] = "Inheritance is enabled."
    else:
        ret["result"] = False
        ret["comment"] = tRet["comment"]
    return ret


def disinherit(name, objectType, copy_inherited_acl=True):
    """
    Ensure an object is not inheriting ACLs from its parent
    """
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}
    tRet = __salt__["win_dacl.check_inheritance"](name, objectType)
    if tRet["result"]:
        if tRet["Inheritance"]:
            if __opts__["test"]:
                ret["result"] = None
                ret["changes"]["Inheritance"] = "Disabled"
                ret["comment"] = "Inheritance is set to be disabled."
                ret["changes"]["Inherited ACLs"] = (
                    "Are set to be kept"
                    if copy_inherited_acl
                    else "Are set to be removed"
                )
                return ret
            eRet = __salt__["win_dacl.disable_inheritance"](
                name, objectType, copy_inherited_acl
            )
            ret["result"] = eRet["result"]
            if eRet["result"]:
                ret["changes"] = dict(ret["changes"], **eRet["changes"])
            else:
                ret["comment"] = " ".join([ret["comment"], eRet["comment"]])
        else:
            if __opts__["test"]:
                ret["result"] = None
                ret["comment"] = "Inheritance is disabled."
    else:
        ret["result"] = False
        ret["comment"] = tRet["comment"]
    return ret



import pytest
from tornado.httpclient import HTTPError

import salt.utils.json
from salt.netapi.rest_tornado import saltnado


@pytest.fixture
def app_urls():
    return [
        (r"/minions/(.*)", saltnado.MinionSaltAPIHandler),
        (r"/minions", saltnado.MinionSaltAPIHandler),
    ]


async def test_get_no_mid(http_client, salt_minion, salt_sub_minion):
    response = await http_client.fetch(
        "/minions",
        method="GET",
        follow_redirects=False,
    )
    print(f"{response!r}")
    response_obj = salt.utils.json.loads(response.body)
    assert len(response_obj["return"]) == 1
    assert isinstance(response_obj["return"][0], dict)
    # one per minion
    assert len(response_obj["return"][0]) == 2
    assert salt_minion.id in response_obj["return"][0]
    assert salt_sub_minion.id in response_obj["return"][0]
    # check a single grain
    for minion_id, grains in response_obj["return"][0].items():
        assert minion_id == grains["id"]


@pytest.mark.slow_test
async def test_get(http_client, salt_minion):
    response = await http_client.fetch(
        f"/minions/{salt_minion.id}",
        method="GET",
        follow_redirects=False,
    )
    response_obj = salt.utils.json.loads(response.body)
    assert len(response_obj["return"]) == 1
    assert isinstance(response_obj["return"][0], dict)
    assert len(response_obj["return"][0]) == 1
    assert salt_minion.id in response_obj["return"][0]
    # check a single grain
    assert response_obj["return"][0][salt_minion.id]["id"] == salt_minion.id


async def test_post(http_client, salt_minion, salt_sub_minion):
    low = [{"tgt": "*minion-*", "fun": "test.ping"}]
    response = await http_client.fetch(
        "/minions",
        method="POST",
        body=salt.utils.json.dumps(low),
    )

    response_obj = salt.utils.json.loads(response.body)
    ret = response_obj["return"]

    # TODO: verify pub function? Maybe look at how we test the publisher
    assert len(ret) == 1
    assert "jid" in ret[0]
    assert sorted(ret[0]["minions"]) == sorted([salt_minion.id, salt_sub_minion.id])


@pytest.mark.slow_test
async def test_post_with_client(http_client, salt_minion, salt_sub_minion):
    # get a token for this test
    low = [{"client": "local_async", "tgt": "*minion-*", "fun": "test.ping"}]
    response = await http_client.fetch(
        "/minions",
        method="POST",
        body=salt.utils.json.dumps(low),
    )

    response_obj = salt.utils.json.loads(response.body)
    ret = response_obj["return"]

    # TODO: verify pub function? Maybe look at how we test the publisher
    assert len(ret) == 1
    assert "jid" in ret[0]
    assert sorted(ret[0]["minions"]) == sorted([salt_minion.id, salt_sub_minion.id])


@pytest.mark.slow_test
async def test_post_with_incorrect_client(http_client):
    """
    The /minions endpoint is asynchronous only, so if you try something else
    make sure you get an error
    """
    # get a token for this test
    low = [{"client": "local", "tgt": "*", "fun": "test.ping"}]
    with pytest.raises(HTTPError) as exc:
        await http_client.fetch(
            "/minions",
            method="POST",
            body=salt.utils.json.dumps(low),
        )
    assert exc.value.code == 400


@pytest.mark.slow_test
async def test_mem_leak_in_event_listener(http_client, salt_minion, app):
    for i in range(10):
        await http_client.fetch(
            f"/minions/{salt_minion.id}",
            method="GET",
            follow_redirects=False,
        )
    assert len(app.event_listener.tag_map) == 0
    assert len(app.event_listener.timeout_map) == 0
    assert len(app.event_listener.request_map) == 0



"""
tests.unit.utils.test_configparser
==================================

Test the funcs in the custom parsers in salt.utils.configparser
"""

import copy
import errno
import logging
import os

import salt.utils.configparser
import salt.utils.files
import salt.utils.platform
import salt.utils.stringutils
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


# The user.name param here is intentionally indented with spaces instead of a
# tab to test that we properly load a file with mixed indentation.
ORIG_CONFIG = """[user]
        name = Артём Анисимов
\temail = foo@bar.com
[remote "origin"]
\turl = https://github.com/terminalmage/salt.git
\tfetch = +refs/heads/*:refs/remotes/origin/*
\tpushurl = git@github.com:terminalmage/salt.git
[color "diff"]
\told = 196
\tnew = 39
[core]
\tpager = less -R
\trepositoryformatversion = 0
\tfilemode = true
\tbare = false
\tlogallrefupdates = true
[alias]
\tmodified = ! git status --porcelain | awk 'match($1, "M"){print $2}'
\tgraph = log --all --decorate --oneline --graph
\thist = log --pretty=format:\\"%h %ad | %s%d [%an]\\" --graph --date=short
[http]
\tsslverify = false""".split(
    "\n"
)


class TestGitConfigParser(TestCase):
    """
    Tests for salt.utils.configparser.GitConfigParser
    """

    maxDiff = None
    orig_config = os.path.join(RUNTIME_VARS.TMP, "test_gitconfig.orig")
    new_config = os.path.join(RUNTIME_VARS.TMP, "test_gitconfig.new")
    remote = 'remote "origin"'

    def tearDown(self):
        del self.conf
        try:
            os.remove(self.new_config)
        except OSError as exc:
            if exc.errno != errno.ENOENT:
                raise

    def setUp(self):
        if not os.path.exists(self.orig_config):
            with salt.utils.files.fopen(self.orig_config, "wb") as fp_:
                fp_.write(salt.utils.stringutils.to_bytes(os.linesep.join(ORIG_CONFIG)))
        self.conf = salt.utils.configparser.GitConfigParser()
        with salt.utils.files.fopen(self.orig_config, "rb") as fp:
            self.conf._read(fp, self.orig_config)

    @classmethod
    def tearDownClass(cls):
        try:
            os.remove(cls.orig_config)
        except OSError as exc:
            if exc.errno != errno.ENOENT:
                raise

    @staticmethod
    def fix_indent(lines):
        """
        Fixes the space-indented 'user' line, because when we write the config
        object to a file space indentation will be replaced by tab indentation.
        """
        ret = copy.copy(lines)
        for i, _ in enumerate(ret):
            if ret[i].startswith(salt.utils.configparser.GitConfigParser.SPACEINDENT):
                ret[i] = ret[i].replace(
                    salt.utils.configparser.GitConfigParser.SPACEINDENT, "\t"
                )
        return ret

    @staticmethod
    def get_lines(path):
        with salt.utils.files.fopen(path, "rb") as fp_:
            return salt.utils.stringutils.to_unicode(fp_.read()).splitlines()

    def _test_write(self, mode):
        kwargs = {"mode": mode}
        if salt.utils.platform.is_windows() and "b" not in mode:
            kwargs["encoding"] = "utf-8"
        with salt.utils.files.fopen(self.new_config, **kwargs) as fp_:
            self.conf.write(fp_)
        self.assertEqual(self.get_lines(self.new_config), self.fix_indent(ORIG_CONFIG))

    def test_get(self):
        """
        Test getting an option's value
        """
        # Numeric values should be loaded as strings
        self.assertEqual(self.conf.get('color "diff"', "old"), "196")
        # Complex strings should be loaded with their literal quotes and
        # slashes intact
        self.assertEqual(
            self.conf.get("alias", "modified"),
            """! git status --porcelain | awk 'match($1, "M"){print $2}'""",
        )
        self.assertEqual(
            self.conf.get("alias", "hist"),
            salt.utils.stringutils.to_unicode(
                r"""log --pretty=format:\"%h %ad | %s%d [%an]\" --graph --date=short"""
            ),
        )

    def test_read_space_indent(self):
        """
        Test that user.name was successfully loaded despite being indented
        using spaces instead of a tab. Additionally, this tests that the value
        was loaded as a unicode type on PY2.
        """
        self.assertEqual(self.conf.get("user", "name"), "Артём Анисимов")

    def test_set_new_option(self):
        """
        Test setting a new option in an existing section
        """
        self.conf.set("http", "useragent", "myawesomeagent")
        self.assertEqual(self.conf.get("http", "useragent"), "myawesomeagent")

    def test_add_section(self):
        """
        Test adding a section and adding an item to that section
        """
        self.conf.add_section("foo")
        self.conf.set("foo", "bar", "baz")
        self.assertEqual(self.conf.get("foo", "bar"), "baz")

    def test_replace_option(self):
        """
        Test replacing an existing option
        """
        # We're also testing the normalization of key names, here. Setting
        # "sslVerify" should actually set an "sslverify" option.
        self.conf.set("http", "sslVerify", "true")
        self.assertEqual(self.conf.get("http", "sslverify"), "true")

    def test_set_multivar(self):
        """
        Test setting a multivar and then writing the resulting file
        """
        orig_refspec = "+refs/heads/*:refs/remotes/origin/*"
        new_refspec = "+refs/tags/*:refs/tags/*"
        # Make sure that the original value is a string
        self.assertEqual(self.conf.get(self.remote, "fetch"), orig_refspec)
        # Add another refspec
        self.conf.set_multivar(self.remote, "fetch", new_refspec)
        # The value should now be a list
        self.assertEqual(
            self.conf.get(self.remote, "fetch"), [orig_refspec, new_refspec]
        )
        # Write the config object to a file
        with salt.utils.files.fopen(self.new_config, "wb") as fp_:
            self.conf.write(fp_)
        # Confirm that the new file was written correctly
        expected = self.fix_indent(ORIG_CONFIG)
        expected.insert(6, f"\tfetch = {new_refspec}")
        self.assertEqual(self.get_lines(self.new_config), expected)

    def test_remove_option(self):
        """
        test removing an option, including all items from a multivar
        """
        for item in ("fetch", "pushurl"):
            self.conf.remove_option(self.remote, item)
            # To confirm that the option is now gone, a get should raise an
            # NoOptionError exception.
            self.assertRaises(
                salt.utils.configparser.NoOptionError, self.conf.get, self.remote, item
            )

    def test_remove_option_regexp(self):
        """
        test removing an option, including all items from a multivar
        """
        orig_refspec = "+refs/heads/*:refs/remotes/origin/*"
        new_refspec_1 = "+refs/tags/*:refs/tags/*"
        new_refspec_2 = "+refs/foo/*:refs/foo/*"
        # First, add both refspecs
        self.conf.set_multivar(self.remote, "fetch", new_refspec_1)
        self.conf.set_multivar(self.remote, "fetch", new_refspec_2)
        # Make sure that all three values are there
        self.assertEqual(
            self.conf.get(self.remote, "fetch"),
            [orig_refspec, new_refspec_1, new_refspec_2],
        )
        # If the regex doesn't match, no items should be removed
        self.assertFalse(
            self.conf.remove_option_regexp(
                self.remote,
                "fetch",
                salt.utils.stringutils.to_unicode(r"\d{7,10}"),
            )
        )
        # Make sure that all three values are still there (since none should
        # have been removed)
        self.assertEqual(
            self.conf.get(self.remote, "fetch"),
            [orig_refspec, new_refspec_1, new_refspec_2],
        )
        # Remove one of the values
        self.assertTrue(self.conf.remove_option_regexp(self.remote, "fetch", "tags"))
        # Confirm that the value is gone
        self.assertEqual(
            self.conf.get(self.remote, "fetch"), [orig_refspec, new_refspec_2]
        )
        # Remove the other one we added earlier
        self.assertTrue(self.conf.remove_option_regexp(self.remote, "fetch", "foo"))
        # Since the option now only has one value, it should be a string
        self.assertEqual(self.conf.get(self.remote, "fetch"), orig_refspec)
        # Remove the last remaining option
        self.assertTrue(self.conf.remove_option_regexp(self.remote, "fetch", "heads"))
        # Trying to do a get now should raise an exception
        self.assertRaises(
            salt.utils.configparser.NoOptionError, self.conf.get, self.remote, "fetch"
        )

    def test_write(self):
        """
        Test writing using non-binary filehandle
        """
        self._test_write(mode="w")

    def test_write_binary(self):
        """
        Test writing using binary filehandle
        """
        self._test_write(mode="wb")



"""
Manage PagerDuty users.

Example:

    .. code-block:: yaml

    ensure bruce test user 1:
        pagerduty.user_present:
            - name: 'Bruce TestUser1'
            - email: bruce+test1@lyft.com
            - requester_id: P1GV5NT

"""


def __virtual__():
    """
    Only load if the pygerduty module is available in __salt__
    """
    if "pagerduty_util.get_resource" in __salt__:
        return "pagerduty_user"
    return (False, "pagerduty_util module could not be loaded")


def present(profile="pagerduty", subdomain=None, api_key=None, **kwargs):
    """
    Ensure pagerduty user exists.
    Arguments match those supported by
    https://developer.pagerduty.com/documentation/rest/users/create.
    """
    return __salt__["pagerduty_util.resource_present"](
        "users", ["email", "name", "id"], None, profile, subdomain, api_key, **kwargs
    )


def absent(profile="pagerduty", subdomain=None, api_key=None, **kwargs):
    """
    Ensure pagerduty user does not exist.
    Name can be pagerduty id, email address, or user name.
    """
    return __salt__["pagerduty_util.resource_absent"](
        "users", ["email", "name", "id"], profile, subdomain, api_key, **kwargs
    )



"""
CouchDB sdb Module

:maintainer:    SaltStack
:maturity:      New
:depends:       python2-couchdb
:platform:      all

This allow interaction between Salt and a CouchDB [couchdb.apache.org]
database. It uses salt's `sdb` system to allow for inserts and retrevals
using the `sdb://` prefix in salt configuration files.

To use the couchbase sdb module, it must first be configured in the salt
master or minion config. The following arguments are required:

.. code-block:: yaml

    couchdb_sdb:
      driver: couchdb
      host: localhost
      port: 5984
      database: salt_sdb

One could then query the CouchDB instance via an `sdb://` URI such as the
following:

.. code-block:: yaml

    password: sdb://couchdb_sdb/mykey

To use this interface, you must track IDs on your own or have another source
to do the map-reduce logic necessary to calculate the ID you wish to fetch.

Additional contributions to build true map-reduce functionality into this module
would be welcome.
"""

# Import Python libraries
import logging
from uuid import uuid4

# Import Salt libraries
from salt.utils.decorators import memoize

try:
    import couchdb

    HAS_COUCH = True
except ImportError:
    HAS_COUCH = False


log = logging.getLogger(__name__)

# 'set' is a reserved word
__func_alias__ = {"set_": "set"}


def __virtual__():
    """
    Require the python2-couchdb libraries
    """
    return HAS_COUCH


@memoize
def _construct_uri(profile):
    """
    Examine configuration and return
    a uri for the couchdb server in the following format:

    .. code-block:: bash

        http://localhost:5984/
    """
    return "http://{host}:{port}".format(**profile)


def _get_conn(profile):
    """
    Get a connection to CouchDB
    """
    DEFAULT_BASE_URL = _construct_uri(profile) or "http://localhost:5984"

    server = couchdb.Server()
    if profile["database"] not in server:
        server.create(profile["database"])
    return server


def set_(key, value, profile=None):
    """
    Set a key/value pair in couchdb
    """
    db = _get_db(profile)
    return db.save({"_id": uuid4().hex, key: value})


def get(key, profile=None):
    """
    Get a value from couchdb by id
    """
    db = _get_db(profile)
    return db.get(key)


def _get_db(profile):
    """
    Wraps _get_conn() to return a db
    """
    server = _get_conn(profile)
    db = _get_db(profile)
    return db
